# -*- coding: utf-8 -*-

"""Define module level fixtures."""


from __future__ import absolute_import

import pytest
from pandas import Series

import cobra.util.solver as sutil
from cobra.core import Metabolite, Model, Reaction, Solution


# The scipy interface is currently unstable and may yield errors or infeasible
# solutions.


@pytest.fixture(scope="session",
                params=[s for s in ["glpk", "cplex", "gurobi"]
                        if s in sutil.solvers])
def all_solvers(request):
    """Return the avaialble solvers."""
    return request.param


@pytest.fixture(scope="session",
                params=[s for s in ["cplex", "gurobi"]
                        if s in sutil.solvers])
def qp_solvers(request):
    """Return the available QP solvers."""
    return request.param


@pytest.fixture(scope="module")
def room_model():
    """
    Generate ROOM model as described in [1]_

    References
    ----------
    .. [1] Tomer Shlomi, Omer Berkman and Eytan Ruppin, "Regulatory on/off
     minimization of metabolic flux changes after genetic perturbations",
     PNAS 2005 102 (21) 7695-7700; doi:10.1073/pnas.0406346102

    """
    test_model = Model("papin_2003")

    v_1 = Reaction("v1")
    v_2 = Reaction("v2")
    v_3 = Reaction("v3")
    v_4 = Reaction("v4")
    v_5 = Reaction("v5")
    v_6 = Reaction("v6", upper_bound=0.0)
    b_1 = Reaction("b1", upper_bound=10.0, lower_bound=0.0)
    b_2 = Reaction("b2")
    b_3 = Reaction("b3")

    test_model.add_reactions([v_1, v_2, v_3, v_4, v_5, v_6, b_1, b_2, b_3])

    v_1.reaction = "A -> B"
    v_2.reaction = "2 B -> C + byp"
    v_3.reaction = "2 B + cof -> D"
    v_4.reaction = "D -> E + cof"
    v_5.reaction = "C + cof -> D"
    v_6.reaction = "C -> E"
    b_1.reaction = "-> A"
    b_2.reaction = "E ->"
    b_3.reaction = "byp ->"

    test_model.objective = 'b2'

    return test_model


@pytest.fixture(scope="module")
def room_solution():
    """
    Generate ROOM solution as described in [1]_

    References
    ----------
    .. [1] Tomer Shlomi, Omer Berkman and Eytan Ruppin, "Regulatory on/off
     minimization of metabolic flux changes after genetic perturbations",
     PNAS 2005 102 (21) 7695-7700; doi:10.1073/pnas.0406346102

    """
    fluxes = Series({'b1': 10.0, 'b2': 5.0, 'b3': 5.0, 'v1': 10.0, 'v2': 5.0,
                     'v3': 0.0, 'v4': 0.0, 'v5': 0.0, 'v6': 5.0})
    reduced_costs = Series({'b1': 0.0, 'b2': 0.0, 'b3': 0.0, 'v1': 0.0,
                            'v2': 0.0, 'v3': 0.0, 'v4': 0.0, 'v5': 0.0,
                            'v6': 0.0})
    shadow_prices = Series({'b1': 0.0, 'b2': 0.0, 'b3': 0.0, 'v1': 0.0,
                            'v2': 0.0, 'v3': 0.0, 'v4': 0.0, 'v5': 0.0,
                            'v6': 0.0})
    sol = Solution(objective_value=5.000, status='optimal',
                   fluxes=fluxes,
                   reduced_costs=reduced_costs,
                   shadow_prices=shadow_prices)
    return sol



# -*- coding: utf-8 -*-

"""Test functionalities of flux sampling methods."""

from __future__ import absolute_import

import numpy as np
import pytest

from cobra.core import Metabolite, Model, Reaction
from cobra.flux_analysis.parsimonious import pfba
from cobra.flux_analysis.sampling import ACHRSampler, OptGPSampler, sample


@pytest.fixture(scope="function")
def achr(model):
    """Return ACHRSampler instance for tests."""
    sampler = ACHRSampler(model, thinning=1)
    assert ((sampler.n_warmup > 0) and
            (sampler.n_warmup <= 2 * len(model.variables)))
    assert all(sampler.validate(sampler.warmup) == "v")
    return sampler


@pytest.fixture(scope="function")
def optgp(model):
    """Return OptGPSampler instance for tests."""
    sampler = OptGPSampler(model, processes=1, thinning=1)
    assert ((sampler.n_warmup > 0) and
            (sampler.n_warmup <= 2 * len(model.variables)))
    assert all(sampler.validate(sampler.warmup) == "v")
    return sampler


# Benchmarks
def test_achr_init_benchmark(model, benchmark):
    """Benchmark inital ACHR sampling."""
    benchmark(lambda: ACHRSampler(model))


def test_optgp_init_benchmark(model, benchmark):
    """Benchmark inital OptGP sampling."""
    benchmark(lambda: OptGPSampler(model, processes=2))


def test_achr_sample_benchmark(achr, benchmark):
    """Benchmark ACHR sampling."""
    benchmark(achr.sample, 1)


def test_optgp_sample_benchmark(optgp, benchmark):
    """Benchmark OptGP sampling."""
    benchmark(optgp.sample, 1)


# Tests
def test_single_achr(model):
    """Test ACHR sampling (one sample)."""
    s = sample(model, 10, method="achr")
    assert s.shape == (10, len(model.reactions))


def test_single_optgp(model):
    """Test OptGP sampling (one sample)."""
    s = sample(model, 10, processes=1)
    assert s.shape == (10, len(model.reactions))


def test_multi_optgp(model):
    """Test OptGP sampling (multi sample)."""
    s = sample(model, 10, processes=2)
    assert s.shape == (10, len(model.reactions))


def test_wrong_method(model):
    """Test method intake sanity."""
    with pytest.raises(ValueError):
        sample(model, 1, method="schwupdiwupp")


def test_validate_wrong_sample(achr, model):
    """Test sample correctness."""
    s = achr.sample(10)
    s["hello"] = 1
    with pytest.raises(ValueError):
        achr.validate(s)


def test_fixed_seed(model):
    """Test result of fixed seed for sampling."""
    s1 = sample(model, 1, seed=42)
    s2 = sample(model, 1, seed=42)
    assert np.isclose(s1.TPI[0], s2.TPI[0])


def test_equality_constraint(model):
    """Test equality constraint."""
    model.reactions.ACALD.bounds = (-1.5, -1.5)
    s = sample(model, 10)
    assert np.allclose(s.ACALD, -1.5, atol=1e-6, rtol=0)
    s = sample(model, 10, method="achr")
    assert np.allclose(s.ACALD, -1.5, atol=1e-6, rtol=0)


def test_inequality_constraint(model):
    """Test inequality constraint."""
    co = model.problem.Constraint(
        model.reactions.ACALD.flux_expression, lb=-0.5)
    model.add_cons_vars(co)
    s = sample(model, 10)
    assert all(s.ACALD > -0.5 - 1e-6)
    s = sample(model, 10, method="achr")
    assert all(s.ACALD > -0.5 - 1e-6)


def test_sampling(achr, optgp):
    """Test sampling."""
    s = achr.sample(10)
    assert all(achr.validate(s) == "v")

    s = optgp.sample(10)
    assert all(optgp.validate(s) == "v")


def test_batch_sampling(achr, optgp):
    """Test batch sampling."""
    for b in achr.batch(5, 4):
        assert all(achr.validate(b) == "v")

    for b in optgp.batch(5, 4):
        assert all(optgp.validate(b) == "v")


def test_variables_samples(achr, optgp):
    """Test variable samples."""
    vnames = np.array([v.name for v in achr.model.variables])
    s = achr.sample(10, fluxes=False)
    assert s.shape == (10, achr.warmup.shape[1])
    assert (s.columns == vnames).all()
    assert (achr.validate(s) == "v").all()

    s = optgp.sample(10, fluxes=False)
    assert s.shape == (10, optgp.warmup.shape[1])
    assert (s.columns == vnames).all()
    assert (optgp.validate(s) == "v").all()


def test_inhomogeneous_sanity(model):
    """Test whether inhomogeneous sampling gives approximately the same
    standard deviation as a homogeneous version."""
    model.reactions.ACALD.bounds = (-1.5, -1.5)
    s_inhom = sample(model, 64)

    model.reactions.ACALD.bounds = (-1.5 - 1e-3, -1.5 + 1e-3)
    s_hom = sample(model, 64)

    relative_diff = (s_inhom.std() + 1e-12) / (s_hom.std() + 1e-12)
    assert 0.5 < relative_diff.abs().mean() < 2

    model.reactions.ACALD.bounds = (-1.5, -1.5)
    s_inhom = sample(model, 64, method="achr")

    model.reactions.ACALD.bounds = (-1.5 - 1e-3, -1.5 + 1e-3)
    s_hom = sample(model, 64, method="achr")

    relative_diff = (s_inhom.std() + 1e-12) / (s_hom.std() + 1e-12)
    assert 0.5 < relative_diff.abs().mean() < 2


def test_reproject(optgp):
    """Test reprojection of sampling."""
    s = optgp.sample(10, fluxes=False).values
    proj = np.apply_along_axis(optgp._reproject, 1, s)
    assert all(optgp.validate(proj) == "v")

    s = np.random.rand(10, optgp.warmup.shape[1])
    proj = np.apply_along_axis(optgp._reproject, 1, s)
    assert all(optgp.validate(proj) == "v")


def test_complicated_model():
    """Test a complicated model.

    Difficult model since the online mean calculation is numerically
    unstable so many samples weakly violate the equality constraints.

    """
    model = Model('flux_split')

    reaction1 = Reaction('V1')
    reaction2 = Reaction('V2')
    reaction3 = Reaction('V3')
    reaction1.bounds = (0, 6)
    reaction2.bounds = (0, 8)
    reaction3.bounds = (0, 10)

    A = Metabolite('A')

    reaction1.add_metabolites({A: -1})
    reaction2.add_metabolites({A: -1})
    reaction3.add_metabolites({A: 1})

    model.add_reactions([reaction1, reaction2, reaction3])

    optgp = OptGPSampler(model, 1, seed=42)
    achr = ACHRSampler(model, seed=42)

    optgp_samples = optgp.sample(100)
    achr_samples = achr.sample(100)

    assert any(optgp_samples.corr().abs() < 1.0)
    assert any(achr_samples.corr().abs() < 1.0)
    # > 95% are valid
    assert sum(optgp.validate(optgp_samples) == "v") > 95
    assert sum(achr.validate(achr_samples) == "v") > 95


def test_single_point_space(model):
    """Test the reduction of the sampling space to one point."""
    pfba_sol = pfba(model)
    pfba_const = model.problem.Constraint(
        sum(model.variables), ub=pfba_sol.objective_value)
    model.add_cons_vars(pfba_const)
    model.reactions.Biomass_Ecoli_core.lower_bound = \
        pfba_sol.fluxes.Biomass_Ecoli_core
    with pytest.raises(ValueError):
        s = sample(model, 1)



# -*- coding: utf-8 -*-

"""Test functions of solution.py"""

from __future__ import absolute_import

from cobra.core import Solution


def test_solution_contains_only_reaction_specific_values(solved_model):
    solution, model = solved_model
    reaction_ids = set([reaction.id for reaction in model.reactions])
    if isinstance(solution, Solution):
        assert set(solution.fluxes.index) == reaction_ids
        # assert set(solution.reduced_costs.index) == reaction_ids
    else:
        raise TypeError(
            "solutions of type {0:r} are untested".format(type(solution)))



# -*- coding: utf-8 -*-

"""Test functions of array.py"""

from __future__ import absolute_import

import numpy as np
import pytest

from cobra.util import create_stoichiometric_matrix


try:
    import scipy
except ImportError:
    scipy = None


def test_dense_matrix(model):
    S = create_stoichiometric_matrix(model, array_type='dense', dtype=int)
    assert S.dtype == int
    assert np.allclose(S.max(), [59])

    S_df = create_stoichiometric_matrix(
        model, array_type='DataFrame', dtype=int)
    assert S_df.values.dtype == int
    assert np.all(S_df.columns == [r.id for r in model.reactions])
    assert np.all(S_df.index == [m.id for m in model.metabolites])
    assert np.allclose(S_df.values, S)

    S = create_stoichiometric_matrix(model, array_type='dense', dtype=float)
    solution = model.optimize()
    mass_balance = S.dot(solution.fluxes)
    assert np.allclose(mass_balance, 0)


@pytest.mark.skipif(not scipy, reason='Sparse array methods require scipy')
def test_sparse_matrix(model):
    sparse_types = ['dok', 'lil']

    solution = model.optimize()
    for sparse_type in sparse_types:
        S = create_stoichiometric_matrix(model, array_type=sparse_type)
        mass_balance = S.dot(solution.fluxes)
        assert np.allclose(mass_balance, 0)



# -*- coding: utf-8 -*-

"""Test functionalities provided by sbml3.py"""

from __future__ import absolute_import

from os.path import join

import pytest
from six import itervalues

from cobra import io
from cobra.test.test_io.conftest import compare_models


@pytest.fixture(scope="function")
def mini_fbc2_model(data_directory):
    """Return mini_fbc2 model."""
    return io.sbml3.read_sbml_model(join(data_directory, "mini_fbc2.xml"))


# Benchmarks
def test_benchmark_read(data_directory, benchmark):
    """Benchmark SBML read."""
    benchmark(io.sbml3.read_sbml_model, join(data_directory, "mini_fbc2.xml"))


def test_benchmark_write(model, benchmark, tmpdir):
    """Benchmark SBML write."""
    benchmark(io.sbml3.write_sbml_model, model, tmpdir.join("-bench"))


# Tests
def test_sbml3_error(data_directory):
    """Test invalid SBML read."""
    filename = join(data_directory, "invalid0.xml")
    with pytest.raises(io.sbml3.CobraSBMLError):
        io.read_sbml_model(filename)


def test_validate_sbml_model(data_directory):
    """Test validation of SBML."""
    # invalid SBML
    for i in range(3):
        filename = join(data_directory, "invalid{}.xml".format(i))
        _, errors = io.sbml3.validate_sbml_model(filename)
        assert all(len(v) >= 1 for v in itervalues(errors)) is False

    # valid SBML
    filename = join(data_directory, "mini_fbc2.xml")
    _, errors = io.sbml3.validate_sbml_model(filename)
    assert all(len(v) == 0 for v in itervalues(errors))


@pytest.mark.parametrize("sbml_file", ["mini_fbc2.xml", "mini_fbc2.xml.gz",
                                       "mini_fbc2.xml.bz2"])
def test_read_sbml_model(data_directory, mini_model, sbml_file):
    """Test the reading of a model from SBML3."""
    sbml3_model = io.read_sbml_model(join(data_directory, sbml_file))
    assert compare_models(mini_model, sbml3_model) is None


@pytest.mark.parametrize("ext", [".xml", ".xml.gz", ".xml.bz2"])
def test_write_sbml_model(tmpdir, mini_fbc2_model, ext):
    """Test the writing of model to SBML3."""
    output_file = tmpdir.join("mini_fbc2{}".format(ext))
    io.write_sbml_model(mini_fbc2_model, output_file)
    assert output_file.check()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import re
from math import isinf, isnan
from os.path import isfile
from warnings import warn

from six import iteritems

from cobra.core import Configuration, Metabolite, Model, Reaction
from cobra.util.solver import set_objective


try:
    import libsbml
except ImportError:
    libsbml = None


CONFIGURATION = Configuration()


def parse_legacy_id(the_id, the_compartment=None, the_type='metabolite',
                    use_hyphens=False):
    """Deals with a bunch of problems due to bigg.ucsd.edu not following SBML
    standards

    Parameters
    ----------
    the_id: String.
    the_compartment: String
    the_type: String
        Currently only 'metabolite' is supported
    use_hyphens: Boolean
        If True, double underscores (__) in an SBML ID will be converted to
        hyphens

    Returns
    -------
    string: the identifier
    """
    if use_hyphens:
        the_id = the_id.replace('__', '-')
    if the_type == 'metabolite':
        if the_id.split('_')[-1] == the_compartment:
            # Reformat Ids to match convention in Palsson Lab.
            the_id = the_id[:-len(the_compartment) - 1]
        the_id += '[%s]' % the_compartment
    return the_id


def create_cobra_model_from_sbml_file(sbml_filename, old_sbml=False,
                                      legacy_metabolite=False,
                                      print_time=False, use_hyphens=False):
    """convert an SBML XML file into a cobra.Model object.

    Supports SBML Level 2 Versions 1 and 4.  The function will detect if the
    SBML fbc package is used in the file and run the converter if the fbc
    package is used.

    Parameters
    ----------
    sbml_filename: string
    old_sbml: bool
        Set to True if the XML file has metabolite formula appended to
        metabolite names. This was a poorly designed artifact that persists in
        some models.
    legacy_metabolite: bool
        If True then assume that the metabolite id has the compartment id
         appended after an underscore (e.g. _c for cytosol). This has not been
         implemented but will be soon.
    print_time: bool
         deprecated
    use_hyphens: bool
        If True, double underscores (__) in an SBML ID will be converted to
        hyphens

    Returns
    -------
    Model : The parsed cobra model
    """
    if not libsbml:
        raise ImportError('create_cobra_model_from_sbml_file '
                          'requires python-libsbml')

    __default_objective_coefficient = 0
    # Ensure that the file exists
    if not isfile(sbml_filename):
        raise IOError('Your SBML file is not found: %s' % sbml_filename)
    # Expressions to change SBML Ids to Palsson Lab Ids
    metabolite_re = re.compile('^M_')
    reaction_re = re.compile('^R_')
    compartment_re = re.compile('^C_')
    if print_time:
        warn("print_time is deprecated", DeprecationWarning)
    model_doc = libsbml.readSBML(sbml_filename)
    if model_doc.getPlugin("fbc") is not None:
        from libsbml import ConversionProperties, LIBSBML_OPERATION_SUCCESS
        conversion_properties = ConversionProperties()
        conversion_properties.addOption(
            "convert fbc to cobra", True, "Convert FBC model to Cobra model")
        result = model_doc.convert(conversion_properties)
        if result != LIBSBML_OPERATION_SUCCESS:
            raise Exception("Conversion of SBML+fbc to COBRA failed")
    sbml_model = model_doc.getModel()
    sbml_model_id = sbml_model.getId()
    sbml_species = sbml_model.getListOfSpecies()
    sbml_reactions = sbml_model.getListOfReactions()
    sbml_compartments = sbml_model.getListOfCompartments()
    compartment_dict = dict([(compartment_re.split(x.getId())[-1], x.getName())
                             for x in sbml_compartments])
    if legacy_metabolite:
        # Deal with the palsson lab appending the compartment id to the
        # metabolite id
        new_dict = {}
        for the_id, the_name in compartment_dict.items():
            if the_name == '':
                new_dict[the_id[0].lower()] = the_id
            else:
                new_dict[the_id] = the_name
        compartment_dict = new_dict
        legacy_compartment_converter = dict(
            [(v, k) for k, v in iteritems(compartment_dict)])

    cobra_model = Model(sbml_model_id)
    metabolites = []
    metabolite_dict = {}
    # Convert sbml_metabolites to cobra.Metabolites
    for sbml_metabolite in sbml_species:
        # Skip sbml boundary species
        if sbml_metabolite.getBoundaryCondition():
            continue

        if (old_sbml or legacy_metabolite) and \
                sbml_metabolite.getId().endswith('_b'):
            # Deal with incorrect sbml from bigg.ucsd.edu
            continue
        tmp_metabolite = Metabolite()
        metabolite_id = tmp_metabolite.id = sbml_metabolite.getId()
        tmp_metabolite.compartment = compartment_re.split(
            sbml_metabolite.getCompartment())[-1]
        if legacy_metabolite:
            if tmp_metabolite.compartment not in compartment_dict:
                tmp_metabolite.compartment = legacy_compartment_converter[
                    tmp_metabolite.compartment]
            tmp_metabolite.id = parse_legacy_id(
                tmp_metabolite.id, tmp_metabolite.compartment,
                use_hyphens=use_hyphens)
        if use_hyphens:
            tmp_metabolite.id = metabolite_re.split(
                tmp_metabolite.id)[-1].replace('__', '-')
        else:
            # Just in case the SBML ids are ill-formed and use -
            tmp_metabolite.id = metabolite_re.split(
                tmp_metabolite.id)[-1].replace('-', '__')
        tmp_metabolite.name = sbml_metabolite.getName()
        tmp_formula = ''
        tmp_metabolite.notes = parse_legacy_sbml_notes(
            sbml_metabolite.getNotesString())
        if sbml_metabolite.isSetCharge():
            tmp_metabolite.charge = sbml_metabolite.getCharge()
        if "CHARGE" in tmp_metabolite.notes:
            note_charge = tmp_metabolite.notes["CHARGE"][0]
            try:
                note_charge = float(note_charge)
                if note_charge == int(note_charge):
                    note_charge = int(note_charge)
            except:
                warn("charge of %s is not a number (%s)" %
                     (tmp_metabolite.id, str(note_charge)))
            else:
                if ((tmp_metabolite.charge is None) or
                        (tmp_metabolite.charge == note_charge)):
                    tmp_metabolite.notes.pop("CHARGE")
                    # set charge to the one from notes if not assigend before
                    # the same
                    tmp_metabolite.charge = note_charge
                else:  # tmp_metabolite.charge != note_charge
                    msg = "different charges specified for %s (%d and %d)"
                    msg = msg % (tmp_metabolite.id,
                                 tmp_metabolite.charge, note_charge)
                    warn(msg)
                    # Chances are a 0 note charge was written by mistake. We
                    # will default to the note_charge in this case.
                    if tmp_metabolite.charge == 0:
                        tmp_metabolite.charge = note_charge

        for the_key in tmp_metabolite.notes.keys():
            if the_key.lower() == 'formula':
                tmp_formula = tmp_metabolite.notes.pop(the_key)[0]
                break
        if tmp_formula == '' and old_sbml:
            tmp_formula = tmp_metabolite.name.split('_')[-1]
            tmp_metabolite.name = tmp_metabolite.name[:-len(tmp_formula) - 1]
        tmp_metabolite.formula = tmp_formula
        metabolite_dict.update({metabolite_id: tmp_metabolite})
        metabolites.append(tmp_metabolite)
    cobra_model.add_metabolites(metabolites)

    # Construct the vectors and matrices for holding connectivity and numerical
    # info to feed to the cobra toolbox.
    # Always assume steady state simulations so b is set to 0
    cobra_reaction_list = []
    coefficients = {}
    for sbml_reaction in sbml_reactions:
        if use_hyphens:
            # Change the ids to match conventions used by the Palsson lab.
            reaction = Reaction(reaction_re.split(
                sbml_reaction.getId())[-1].replace('__', '-'))
        else:
            # Just in case the SBML ids are ill-formed and use -
            reaction = Reaction(reaction_re.split(
                sbml_reaction.getId())[-1].replace('-', '__'))
        cobra_reaction_list.append(reaction)
        # reaction.exchange_reaction = 0
        reaction.name = sbml_reaction.getName()
        cobra_metabolites = {}
        # Use the cobra.Metabolite class here
        for sbml_metabolite in sbml_reaction.getListOfReactants():
            tmp_metabolite_id = sbml_metabolite.getSpecies()
            # This deals with boundary metabolites
            if tmp_metabolite_id in metabolite_dict:
                tmp_metabolite = metabolite_dict[tmp_metabolite_id]
                cobra_metabolites[tmp_metabolite] = - \
                    sbml_metabolite.getStoichiometry()
        for sbml_metabolite in sbml_reaction.getListOfProducts():
            tmp_metabolite_id = sbml_metabolite.getSpecies()
            # This deals with boundary metabolites
            if tmp_metabolite_id in metabolite_dict:
                tmp_metabolite = metabolite_dict[tmp_metabolite_id]
                # Handle the case where the metabolite was specified both
                # as a reactant and as a product.
                if tmp_metabolite in cobra_metabolites:
                    warn("%s appears as a reactant and product %s" %
                         (tmp_metabolite_id, reaction.id))
                    cobra_metabolites[
                        tmp_metabolite] += sbml_metabolite.getStoichiometry()
                    # if the combined stoichiometry is 0, remove the metabolite
                    if cobra_metabolites[tmp_metabolite] == 0:
                        cobra_metabolites.pop(tmp_metabolite)
                else:
                    cobra_metabolites[
                        tmp_metabolite] = sbml_metabolite.getStoichiometry()
        # check for nan
        for met, v in iteritems(cobra_metabolites):
            if isnan(v) or isinf(v):
                warn("invalid value %s for metabolite '%s' in reaction '%s'" %
                     (str(v), met.id, reaction.id))
        reaction.add_metabolites(cobra_metabolites)
        # Parse the kinetic law info here.
        parameter_dict = {}
        # If lower and upper bounds are specified in the Kinetic Law then
        # they override the sbml reversible attribute.  If they are not
        # specified then the bounds are determined by getReversible.
        if not sbml_reaction.getKineticLaw():

            if sbml_reaction.getReversible():
                parameter_dict['lower_bound'] = CONFIGURATION.lower_bound
                parameter_dict['upper_bound'] = CONFIGURATION.upper_bound
            else:
                # Assume that irreversible reactions only proceed from left to
                # right.
                parameter_dict['lower_bound'] = 0
                parameter_dict['upper_bound'] = CONFIGURATION.upper_bound

            parameter_dict[
                'objective_coefficient'] = __default_objective_coefficient
        else:
            for sbml_parameter in \
                    sbml_reaction.getKineticLaw().getListOfParameters():
                parameter_dict[
                    sbml_parameter.getId().lower()] = sbml_parameter.getValue()

        if 'lower_bound' in parameter_dict:
            reaction.lower_bound = parameter_dict['lower_bound']
        elif 'lower bound' in parameter_dict:
            reaction.lower_bound = parameter_dict['lower bound']
        elif sbml_reaction.getReversible():
            reaction.lower_bound = CONFIGURATION.lower_bound
        else:
            reaction.lower_bound = 0

        if 'upper_bound' in parameter_dict:
            reaction.upper_bound = parameter_dict['upper_bound']
        elif 'upper bound' in parameter_dict:
            reaction.upper_bound = parameter_dict['upper bound']
        else:
            reaction.upper_bound = CONFIGURATION.upper_bound

        objective_coefficient = parameter_dict.get(
            'objective_coefficient', parameter_dict.get(
                'objective_coefficient', __default_objective_coefficient))
        if objective_coefficient != 0:
            coefficients[reaction] = objective_coefficient

        # ensure values are not set to nan or inf
        if isnan(reaction.lower_bound) or isinf(reaction.lower_bound):
            reaction.lower_bound = CONFIGURATION.lower_bound
        if isnan(reaction.upper_bound) or isinf(reaction.upper_bound):
            reaction.upper_bound = CONFIGURATION.upper_bound

        reaction_note_dict = parse_legacy_sbml_notes(
            sbml_reaction.getNotesString())
        # Parse the reaction notes.
        # POTENTIAL BUG: DEALING WITH LEGACY 'SBML' THAT IS NOT IN A
        # STANDARD FORMAT
        # TODO: READ IN OTHER NOTES AND GIVE THEM A reaction_ prefix.
        # TODO: Make sure genes get added as objects
        if 'GENE ASSOCIATION' in reaction_note_dict:
            rule = reaction_note_dict['GENE ASSOCIATION'][0]
            try:
                rule.encode('ascii')
            except (UnicodeEncodeError, UnicodeDecodeError):
                warn("gene_reaction_rule '%s' is not ascii compliant" % rule)
            if rule.startswith("&quot;") and rule.endswith("&quot;"):
                rule = rule[6:-6]
            reaction.gene_reaction_rule = rule
            if 'GENE LIST' in reaction_note_dict:
                reaction.systematic_names = reaction_note_dict['GENE LIST'][0]
            elif ('GENES' in reaction_note_dict and
                  reaction_note_dict['GENES'] != ['']):
                reaction.systematic_names = reaction_note_dict['GENES'][0]
            elif 'LOCUS' in reaction_note_dict:
                gene_id_to_object = dict([(x.id, x) for x in reaction._genes])
                for the_row in reaction_note_dict['LOCUS']:
                    tmp_row_dict = {}
                    the_row = 'LOCUS:' + the_row.lstrip('_').rstrip('#')
                    for the_item in the_row.split('#'):
                        k, v = the_item.split(':')
                        tmp_row_dict[k] = v
                    tmp_locus_id = tmp_row_dict['LOCUS']
                    if 'TRANSCRIPT' in tmp_row_dict:
                        tmp_locus_id = tmp_locus_id + \
                                       '.' + tmp_row_dict['TRANSCRIPT']

                    if 'ABBREVIATION' in tmp_row_dict:
                        gene_id_to_object[tmp_locus_id].name = tmp_row_dict[
                            'ABBREVIATION']

        if 'SUBSYSTEM' in reaction_note_dict:
            reaction.subsystem = reaction_note_dict.pop('SUBSYSTEM')[0]

        reaction.notes = reaction_note_dict

    # Now, add all of the reactions to the model.
    cobra_model.id = sbml_model.getId()
    # Populate the compartment list - This will be done based on
    # cobra.Metabolites in cobra.Reactions in the future.
    cobra_model.compartments = compartment_dict

    cobra_model.add_reactions(cobra_reaction_list)
    set_objective(cobra_model, coefficients)
    return cobra_model


def parse_legacy_sbml_notes(note_string, note_delimiter=':'):
    """Deal with various legacy SBML format issues.
    """
    note_dict = {}
    start_tag = '<p>'
    end_tag = '</p>'
    if '<html:p>' in note_string:
        start_tag = '<html:p>'
        end_tag = '</html:p>'
    while start_tag in note_string and end_tag in note_string:
        note_start = note_string.index(start_tag)
        note_end = note_string.index(end_tag)
        the_note = note_string[
                   (note_start + len(start_tag)):note_end].lstrip(' ').rstrip(
            ' ')
        if note_delimiter in the_note:
            note_delimiter_index = the_note.index(note_delimiter)
            note_field = the_note[:note_delimiter_index].lstrip(
                ' ').rstrip(' ').replace('_', ' ').upper()
            note_value = the_note[
                         (note_delimiter_index + 1):].lstrip(' ').rstrip(' ')
            if note_field in note_dict:
                note_dict[note_field].append(note_value)
            else:
                note_dict[note_field] = [note_value]
        note_string = note_string[(note_end + len(end_tag)):]

    if ('CHARGE' in note_dict and
            note_dict['CHARGE'][0].lower() in ['none', 'na', 'nan']):
        note_dict.pop('CHARGE')  # Remove non-numeric charges

    if 'CHARGE' in note_dict and note_dict['CHARGE'][0].lower() in ['none',
                                                                    'na',
                                                                    'nan']:
        note_dict.pop('CHARGE')  # Remove non-numeric charges

    return note_dict


def write_cobra_model_to_sbml_file(cobra_model, sbml_filename,
                                   sbml_level=2, sbml_version=1,
                                   print_time=False,
                                   use_fbc_package=True):
    """Write a cobra.Model object to an SBML XML file.

    Parameters
    ----------
    cobra_model : cobra.core.Model.Model
        The model object to write
    sbml_filename : string
        The file to write the SBML XML to.
    sbml_level : int
        2 is the only supported level.
    sbml_version : int
        1 is the only supported version.
    print_time : bool
        deprecated
    use_fbc_package : bool
        Convert the model to the FBC package format to improve portability.
        http://sbml.org/Documents/Specifications/SBML_Level_3/Packages/Flux_Balance_Constraints_(flux)

    Notes
    -----
    TODO: Update the NOTES to match the SBML standard and provide support for
    Level 2 Version 4
    """
    if not libsbml:
        raise ImportError('write_cobra_model_to_sbml_file '
                          'requires python-libsbml')
    sbml_doc = get_libsbml_document(cobra_model,
                                    sbml_level=sbml_level,
                                    sbml_version=sbml_version,
                                    print_time=print_time,
                                    use_fbc_package=use_fbc_package)

    libsbml.writeSBML(sbml_doc, sbml_filename)


def get_libsbml_document(cobra_model,
                         sbml_level=2, sbml_version=1,
                         print_time=False,
                         use_fbc_package=True):
    """ Return a libsbml document object for writing to a file. This function
    is used by write_cobra_model_to_sbml_file(). """
    note_start_tag, note_end_tag = '<p>', '</p>'
    if sbml_level > 2 or (sbml_level == 2 and sbml_version == 4):
        note_start_tag, note_end_tag = '<html:p>', '</html:p>'

    sbml_doc = libsbml.SBMLDocument(sbml_level, sbml_version)
    sbml_model = sbml_doc.createModel(cobra_model.id.split('.')[0])
    # Note need to set units
    reaction_units = 'mmol_per_gDW_per_hr'
    model_units = sbml_model.createUnitDefinition()
    model_units.setId(reaction_units)
    sbml_unit = model_units.createUnit()
    sbml_unit.setKind(libsbml.UNIT_KIND_MOLE)
    sbml_unit.setScale(-3)
    sbml_unit = model_units.createUnit()
    sbml_unit.setKind(libsbml.UNIT_KIND_GRAM)
    sbml_unit.setExponent(-1)
    sbml_unit = model_units.createUnit()
    sbml_unit.setKind(libsbml.UNIT_KIND_SECOND)
    sbml_unit.setMultiplier(1.0 / 60 / 60)
    sbml_unit.setExponent(-1)

    # Add in the common compartment abbreviations.  If there are additional
    # compartments they also need to be added.
    if not cobra_model.compartments:
        cobra_model.compartments = {'c': 'cytosol',
                                    'p': 'periplasm',
                                    'e': 'extracellular'}
    for the_key in cobra_model.compartments.keys():
        sbml_comp = sbml_model.createCompartment()
        sbml_comp.setId(the_key)
        sbml_comp.setName(cobra_model.compartments[the_key])
        sbml_comp.setSize(1)  # Just to get rid of warnings

    if print_time:
        warn("print_time is deprecated", DeprecationWarning)
    # Use this dict to allow for fast look up of species id
    # for references created in the reaction section.
    metabolite_dict = {}

    for cobra_metabolite in cobra_model.metabolites:
        metabolite_dict[cobra_metabolite.id] = add_sbml_species(
            sbml_model, cobra_metabolite, note_start_tag=note_start_tag,
            note_end_tag=note_end_tag)

    for the_reaction in cobra_model.reactions:
        # This is probably the culprit.  Including cobra.Reaction
        # objects explicitly in cobra.Model will speed this up.
        sbml_reaction = sbml_model.createReaction()
        # Need to remove - for proper SBML.  Replace with __
        the_reaction_id = 'R_' + the_reaction.id.replace('-', '__')
        sbml_reaction.setId(the_reaction_id)
        # The reason we are not using the Reaction.reversibility property
        # is because the SBML definition of reversibility does not quite
        # match with the cobra definition. In cobra, reversibility implies
        # that both positive and negative flux values are feasible. However,
        # SBML requires negative-flux-only reactions to still be classified
        # as reversible. To quote from the SBML Level 3 Version 1 Spec:
        # > However, labeling a reaction as irreversible is interpreted as
        # > an assertion that the rate expression will not have negative
        # > values during a simulation.
        # (Page 60 lines 44-45)
        sbml_reaction.setReversible(the_reaction.lower_bound < 0)
        if the_reaction.name:
            sbml_reaction.setName(the_reaction.name)
        else:
            sbml_reaction.setName(the_reaction.id)
        # Add in the reactant/product references
        for the_metabolite, the_coefficient in \
                iteritems(the_reaction._metabolites):
            sbml_stoichiometry = the_coefficient
            metabolite_id = str(metabolite_dict[the_metabolite.id])
            # Each SpeciesReference must have a unique id
            if sbml_stoichiometry < 0:
                species_reference = sbml_reaction.createReactant()
            else:
                species_reference = sbml_reaction.createProduct()
            species_reference.setId(metabolite_id + '_' + the_reaction_id)
            species_reference.setSpecies(metabolite_id)
            species_reference.setStoichiometry(abs(sbml_stoichiometry))
        # Deal with the case where the reaction is a boundary reaction
        if len(the_reaction._metabolites) == 1:
            the_metabolite, the_coefficient = list(
                the_reaction._metabolites.items())[0]
            metabolite_id = add_sbml_species(sbml_model, the_metabolite,
                                             note_start_tag=note_start_tag,
                                             note_end_tag=note_end_tag,
                                             boundary_metabolite=True)
            sbml_stoichiometry = -the_coefficient
            # Each SpeciesReference must have a unique id
            if sbml_stoichiometry < 0:
                species_reference = sbml_reaction.createReactant()
            else:
                species_reference = sbml_reaction.createProduct()
            species_reference.setId(metabolite_id + '_' + the_reaction_id)
            species_reference.setSpecies(metabolite_id)
            species_reference.setStoichiometry(abs(sbml_stoichiometry))

        # Add in the kineticLaw
        sbml_law = libsbml.KineticLaw(sbml_level, sbml_version)
        if hasattr(sbml_law, 'setId'):
            sbml_law.setId('FLUX_VALUE')
        sbml_law.setFormula('FLUX_VALUE')
        reaction_parameter_dict = {
            'LOWER_BOUND': [the_reaction.lower_bound, reaction_units],
            'UPPER_BOUND': [the_reaction.upper_bound, reaction_units],
            'FLUX_VALUE': [0, reaction_units],
            'OBJECTIVE_COEFFICIENT': [the_reaction.objective_coefficient,
                                      'dimensionless']}
        for k, v in reaction_parameter_dict.items():
            sbml_parameter = libsbml.Parameter(sbml_level, sbml_version)
            sbml_parameter.setId(k)
            if hasattr(v, '__iter__'):
                sbml_parameter.setValue(v[0])
                sbml_parameter.setUnits(v[1])
            else:
                sbml_parameter.setValue(v)
            sbml_law.addParameter(sbml_parameter)
        sbml_reaction.setKineticLaw(sbml_law)

        # Checks if GPR and Subsystem annotations are present in the notes
        # section and if they are the same as those in the reaction's
        # gene_reaction_rule/ subsystem attribute. If they are not identical,
        # they are set to be identical
        note_dict = the_reaction.notes.copy()
        if the_reaction.gene_reaction_rule:
            note_dict['GENE ASSOCIATION'] = [
                str(the_reaction.gene_reaction_rule)]
        if the_reaction.subsystem:
            note_dict['SUBSYSTEM'] = [str(the_reaction.subsystem)]

        # In a cobrapy model the notes section is stored as a dictionary. The
        # following section turns the key-value-pairs of the dictionary into a
        # string and replaces recurring symbols so that the string has the i
        # required syntax for an SBML doc.
        note_str = str(list(iteritems(note_dict)))
        note_start_tag, note_end_tag, note_delimiter = '<p>', '</p>', ':'
        note_str = note_str.replace('(\'', note_start_tag)
        note_str = note_str.replace('\']),', note_end_tag)
        note_str = note_str.replace('\',', note_delimiter)
        note_str = note_str.replace('\']', '')
        note_str = note_str.replace('[\'', '')
        note_str = note_str.replace(
            '[', '<html xmlns="http://www.w3.org/1999/xhtml">')
        note_str = note_str.replace(')]', note_end_tag + '</html>')
        sbml_reaction.setNotes(note_str)

    if use_fbc_package:
        try:
            from libsbml import ConversionProperties, LIBSBML_OPERATION_SUCCESS
            conversion_properties = ConversionProperties()
            conversion_properties.addOption(
                "convert cobra", True, "Convert Cobra model")
            result = sbml_doc.convert(conversion_properties)
            if result != LIBSBML_OPERATION_SUCCESS:
                raise Exception("Conversion of COBRA to SBML+fbc failed")
        except Exception as e:
            error_string = 'Error saving as SBML+fbc. %s'
            try:
                # Check whether the FbcExtension is there
                from libsbml import FbcExtension
                error_string = error_string % e
            except ImportError:
                error_string = error_string % "FbcExtension not available in "
                "libsbml. If use_fbc_package == True then libsbml must be "
                "compiled with the fbc extension."
                from libsbml import getLibSBMLDottedVersion
                _sbml_version = getLibSBMLDottedVersion()
                _major, _minor, _patch = map(int, _sbml_version.split('.'))
                if _major < 5 or (_major == 5 and _minor < 8):
                    error_string += "You've got libsbml %s installed. "
                    "You need 5.8.0 or later with the fbc package"

            raise Exception(error_string)
    return sbml_doc


def add_sbml_species(sbml_model, cobra_metabolite, note_start_tag,
                     note_end_tag, boundary_metabolite=False):
    """A helper function for adding cobra metabolites to an sbml model.

    Parameters
    ----------
    sbml_model: sbml_model object

    cobra_metabolite: a cobra.Metabolite object

    note_start_tag: string
       the start tag for parsing cobra notes. this will eventually
       be supplanted when COBRA is worked into sbml.

    note_end_tag: string
       the end tag for parsing cobra notes. this will eventually
       be supplanted when COBRA is worked into sbml.
    boundary_metabolite: bool
       if metabolite boundary condition should be set or not

    Returns
    -------
    string: the created metabolite identifier
    """
    sbml_species = sbml_model.createSpecies()
    the_id = 'M_' + cobra_metabolite.id.replace('-', '__')
    # Deal with legacy naming issues
    the_compartment = cobra_metabolite.compartment
    if the_id.endswith('[%s]' % the_compartment):
        the_id = the_id[:-len('[%s]' % the_compartment)]
    elif not the_id.endswith('_%s' % the_compartment):
        the_id += '_%s' % the_compartment
    if boundary_metabolite:
        the_id += '_boundary'
    sbml_species.setId(the_id)
    metabolite_id = the_id
    if boundary_metabolite:
        sbml_species.setBoundaryCondition(True)
    if cobra_metabolite.name:
        sbml_species.setName(cobra_metabolite.name)
    else:
        sbml_species.setName(cobra_metabolite.id)
    if the_compartment is not None:
        try:
            sbml_species.setCompartment(the_compartment)
        except:
            warn('metabolite failed: ' + the_id)
            return cobra_metabolite
    if cobra_metabolite.charge is not None:
        sbml_species.setCharge(cobra_metabolite.charge)
    # Deal with cases where the formula in the model is not an object but as a
    # string
    if cobra_metabolite.formula or cobra_metabolite.notes:
        tmp_note = '<html xmlns="http://www.w3.org/1999/xhtml">'
        if hasattr(cobra_metabolite.formula, 'id'):
            tmp_note += '%sFORMULA: %s%s' % (note_start_tag,
                                             cobra_metabolite.formula.id,
                                             note_end_tag)
        else:
            tmp_note += '%sFORMULA: %s%s' % (note_start_tag,
                                             cobra_metabolite.formula,
                                             note_end_tag)
        if hasattr(cobra_metabolite.notes, 'items'):
            for the_id_type, the_id in cobra_metabolite.notes.items():
                if the_id_type.lower() == 'charge':
                    # Use of notes['CHARGE'] has been deprecated in favor of
                    # metabolite.charge
                    continue
                if not isinstance(the_id_type, str):
                    the_id_type = str(the_id_type)
                if hasattr(the_id, '__iter__') and len(the_id) == 1:
                    the_id = the_id[0]
                if not isinstance(the_id, str):
                    the_id = str(the_id)
                tmp_note += '%s%s: %s%s' % (note_start_tag,
                                            the_id_type,
                                            the_id, note_end_tag)
        sbml_species.setNotes(tmp_note + '</html>')
    return metabolite_id


def fix_legacy_id(id, use_hyphens=False, fix_compartments=False):
    id = id.replace('_DASH_', '__')
    id = id.replace('_FSLASH_', '/')
    id = id.replace('_BSLASH_', "\\")
    id = id.replace('_LPAREN_', '(')
    id = id.replace('_LSQBKT_', '[')
    id = id.replace('_RSQBKT_', ']')
    id = id.replace('_RPAREN_', ')')
    id = id.replace('_COMMA_', ',')
    id = id.replace('_PERIOD_', '.')
    id = id.replace('_APOS_', "'")
    id = id.replace('&amp;', '&')
    id = id.replace('&lt;', '<')
    id = id.replace('&gt;', '>')
    id = id.replace('&quot;', '"')
    if use_hyphens:
        id = id.replace('__', '-')
    else:
        id = id.replace("-", "__")
    if fix_compartments:
        if len(id) > 2:
            if (id[-3] == "(" and id[-1] == ")") or \
                    (id[-3] == "[" and id[-1] == "]"):
                id = id[:-3] + "_" + id[-2]
    return id


def read_legacy_sbml(filename, use_hyphens=False):
    """read in an sbml file and fix the sbml id's"""
    model = create_cobra_model_from_sbml_file(filename, old_sbml=True)
    for metabolite in model.metabolites:
        metabolite.id = fix_legacy_id(metabolite.id)
    model.metabolites._generate_index()
    for reaction in model.reactions:
        reaction.id = fix_legacy_id(reaction.id)
        if reaction.id.startswith("EX_") and reaction.id.endswith("(e)"):
            reaction.id = reaction.id[:-3] + "_e"
    model.reactions._generate_index()
    # remove boundary metabolites (end in _b and only present in exchanges)
    for metabolite in list(model.metabolites):
        if not metabolite.id.endswith("_b"):
            continue
        if len(metabolite._reaction) == 1:
            if list(metabolite._reaction)[0].id.startswith("EX_"):
                metabolite.remove_from_model()
    model.metabolites._generate_index()
    return model



# -*- coding: utf-8 -*-

from cobra.flux_analysis.deletion import (
    double_gene_deletion, double_reaction_deletion, single_gene_deletion,
    single_reaction_deletion)
from cobra.flux_analysis.gapfilling import gapfill
from cobra.flux_analysis.geometric import geometric_fba
from cobra.flux_analysis.loopless import (loopless_solution, add_loopless)
from cobra.flux_analysis.moma import add_moma, moma
from cobra.flux_analysis.parsimonious import pfba
from cobra.flux_analysis.variability import (
    find_blocked_reactions, find_essential_genes, find_essential_reactions,
    flux_variability_analysis)
from cobra.flux_analysis.phenotype_phase_plane import production_envelope
from cobra.flux_analysis.room import add_room, room
from cobra.flux_analysis.sampling import sample



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Insert a TOML header into the latest release note."""

from __future__ import absolute_import, print_function

import sys
from datetime import date
from glob import glob
from builtins import open
from os.path import join, basename
from shutil import copy


def insert_break(lines, break_pos=9):
    """
    Insert a <!--more--> tag for larger release notes.

    Parameters
    ----------
    lines : list of str
        The content of the release note.
    break_pos : int
        Line number before which a break should approximately be inserted.

    Returns
    -------
    list of str
        The text with the inserted tag or no modification if it was
        sufficiently short.
    """
    def line_filter(line):
        if len(line) == 0:
            return True
        return any(line.startswith(c) for c in "-*+")

    if len(lines) <= break_pos:
        return lines
    newlines = [
        i for i, line in enumerate(lines[break_pos:], start=break_pos)
        if line_filter(line.strip())]
    if len(newlines) > 0:
        break_pos = newlines[0]
    lines.insert(break_pos, "<!--more-->\n")
    return lines


def build_hugo_md(filename, tag, bump):
    """
    Build the markdown release notes for Hugo.

    Inserts the required TOML header with specific values and adds a break
    for long release notes.

    Parameters
    ----------
    filename : str, path
        The release notes file.
    tag : str
        The tag, following semantic versioning, of the current release.
    bump : {"major", "minor", "patch", "alpha", "beta"}
        The type of release.

    """
    header = [
        '+++\n',
        'date = "{}"\n'.format(date.today().isoformat()),
        'title = "{}"\n'.format(tag),
        'author = "The COBRApy Team"\n',
        'release = "{}"\n'.format(bump),
        '+++\n',
        '\n'
    ]
    with open(filename, "r") as file_h:
        content = insert_break(file_h.readlines())
    header.extend(content)
    with open(filename, "w") as file_h:
        file_h.writelines(header)


def intify(filename):
    """
    Turn a release note filename into something sortable.

    Parameters
    ----------
    filename : str
        A release note of expected filename format '<major>.<minor>.<patch>.md'.

    Returns
    -------
    tuple
        A pair of the major and minor versions as integers.

    """
    tmp = filename[:-3].split(".")
    return int(tmp[0]), int(tmp[1])


def find_bump(target, tag):
    """Identify the kind of release by comparing to existing ones."""
    tmp = tag.split(".")
    existing = [intify(basename(f)) for f in glob(join(target, "[0-9]*.md"))]
    latest = max(existing)
    if int(tmp[0]) > latest[0]:
        return "major"
    elif int(tmp[1]) > latest[1]:
        return "minor"
    else:
        return "patch"


def main(argv):
    """
    Identify the release type and create a new target file with TOML header.

    Requires three arguments.

    """
    source, target, tag = argv
    if "a" in tag:
        bump = "alpha"
    if "b" in tag:
        bump = "beta"
    else:
        bump = find_bump(target, tag)
    filename = "{}.md".format(tag)
    destination = copy(join(source, filename), target)
    build_hugo_md(destination, tag, bump)


if __name__ == "__main__":
    if len(sys.argv) != 4:
        print("Usage:\n{} <source dir> <target dir> <tag>"
              "".format(sys.argv[0]))
        sys.exit(2)
    sys.exit(main(sys.argv[1:]))



# -*- coding: utf-8 -*-

"""Test _assess functions in reaction.py"""

from __future__ import absolute_import

from cobra.core import Metabolite, Reaction
from cobra.flux_analysis.reaction import assess


def test_assess(model, all_solvers):
    """Test assess functions."""
    with model:
        assert assess(model, model.reactions.GLCpts,
                      solver=all_solvers) is True
        pyr = model.metabolites.pyr_c
        a = Metabolite('a')
        b = Metabolite('b')
        model.add_metabolites([a, b])
        pyr_a2b = Reaction('pyr_a2b')
        pyr_a2b.add_metabolites({pyr: -1, a: -1, b: 1})
        model.add_reactions([pyr_a2b])
        res = assess(model, pyr_a2b, 0.01, solver=all_solvers)
        expected = {
            'precursors': {a: {'required': 0.01, 'produced': 0.0}},
            'products': {b: {'required': 0.01, 'capacity': 0.0}}}
        assert res == expected



# -*- coding: utf-8 -*-

"""Test functionalities of Phenotype Phase Plane Analysis."""

from __future__ import absolute_import

import numpy as np
import pytest

from cobra.flux_analysis.phenotype_phase_plane import production_envelope


def test_envelope_one(model):
    """Test flux of production envelope."""
    df = production_envelope(model, ["EX_o2_e"])
    assert np.isclose(df["flux_maximum"].sum(), 9.342, atol=1e-3)


def test_envelope_multi_reaction_objective(model):
    """Test production of multiple objectives."""
    obj = {model.reactions.EX_ac_e: 1,
           model.reactions.EX_co2_e: 1}
    with pytest.raises(ValueError):
        production_envelope(model, "EX_o2_e", obj)


@pytest.mark.parametrize("variables, num", [
    (["EX_glc__D_e"], 30),
    (["EX_glc__D_e", "EX_o2_e"], 20),
    (["EX_glc__D_e", "EX_o2_e", "EX_ac_e"], 10)
])
def test_multi_variable_envelope(model, variables, num):
    """Test production of envelope (multiple variable)."""
    df = production_envelope(model, variables, points=num)
    assert len(df) == num ** len(variables)


def test_envelope_two(model):
    """Test production of envelope."""
    df = production_envelope(model, ["EX_glc__D_e", "EX_o2_e"],
                             objective="EX_ac_e")
    assert np.isclose(df["flux_maximum"].sum(), 1737.466, atol=1e-3)
    assert np.isclose(df["carbon_yield_maximum"].sum(), 83.579,
                      atol=1e-3)
    assert np.isclose(df["mass_yield_maximum"].sum(), 82.176,
                      atol=1e-3)



# -*- coding: utf-8 -*-

"""Test functionalities of ROOM."""

from __future__ import absolute_import

import numpy as np

from cobra.flux_analysis.room import add_room


def test_room_sanity(model, all_solvers):
    """Test optimization criterion and optimality for ROOM."""
    model.solver = all_solvers
    sol = model.optimize()
    with model:
        model.reactions.PYK.knock_out()
        knock_sol = model.optimize()

    with model:
        # Internally uses pFBA as reference solution.
        add_room(model)
        model.reactions.PYK.knock_out()
        room_sol = model.optimize()

    with model:
        # Use FBA as reference solution.
        add_room(model, solution=sol)
        model.reactions.PYK.knock_out()
        room_sol_ref = model.optimize()

    flux_change = (sol.fluxes - knock_sol.fluxes).abs().sum()
    flux_change_room = (sol.fluxes - room_sol.fluxes).abs().sum()
    flux_change_room_ref = (sol.fluxes - room_sol_ref.fluxes).abs().sum()
    # Expect the ROOM solution to have smaller flux changes in
    # reactions compared to a normal FBA.
    assert flux_change_room < flux_change or \
        np.isclose(flux_change_room, flux_change, atol=1E-06)
    # Expect the FBA-based reference to have less change in
    # flux distribution.
    assert flux_change_room_ref > flux_change_room or \
        np.isclose(flux_change_room_ref, flux_change_room, atol=1E-06)


def test_linear_room_sanity(model, all_solvers):
    """Test optimization criterion and optimality for linear ROOM."""
    model.solver = all_solvers
    sol = model.optimize()
    with model:
        model.reactions.PYK.knock_out()
        knock_sol = model.optimize()

    with model:
        # Internally uses pFBA as reference solution.
        add_room(model, linear=True)
        model.reactions.PYK.knock_out()
        room_sol = model.optimize()

    with model:
        # Use FBA as reference solution.
        add_room(model, solution=sol, linear=True)
        model.reactions.PYK.knock_out()
        room_sol_ref = model.optimize()

    flux_change = (sol.fluxes - knock_sol.fluxes).abs().sum()
    flux_change_room = (sol.fluxes - room_sol.fluxes).abs().sum()
    flux_change_room_ref = (sol.fluxes - room_sol_ref.fluxes).abs().sum()
    # Expect the ROOM solution to have smaller flux changes in
    # reactions compared to a normal FBA.
    assert flux_change_room < flux_change or \
        np.isclose(flux_change_room, flux_change, atol=1E-06)
    # Expect the FBA-based reference to have less change in
    # flux distribution.
    assert flux_change_room_ref > flux_change_room or \
        np.isclose(flux_change_room_ref, flux_change_room, atol=1E-06)



# -*- coding: utf-8 -*-

"""Test functionalities of reaction and gene deletions."""

from __future__ import absolute_import

import math

import numpy as np
import pytest
from pandas import Series
from six import iteritems

from cobra.flux_analysis.deletion import (
    double_gene_deletion, double_reaction_deletion, single_gene_deletion,
    single_reaction_deletion)
from cobra.flux_analysis.room import add_room


# Single gene deletion FBA
def test_single_gene_deletion_fba_benchmark(model, benchmark,
                                            all_solvers):
    """Benchmark single gene deletion using FBA."""
    model.solver = all_solvers
    benchmark(single_gene_deletion, model)


def test_single_gene_deletion_fba(model, all_solvers):
    """Test single gene deletion using FBA."""
    # expected knockouts for textbook model
    model.solver = all_solvers
    growth_dict = {"b0008": 0.87, "b0114": 0.80, "b0116": 0.78,
                   "b2276": 0.21, "b1779": 0.00}
    result = single_gene_deletion(
        model=model,
        gene_list=list(growth_dict),
        method="fba",
        processes=1
    )["growth"]
    for gene, value in iteritems(growth_dict):
        assert np.isclose(result[frozenset([gene])], value,
                          atol=1E-02)


# Singe gene deletion MOMA
def test_single_gene_deletion_moma_benchmark(model, benchmark,
                                             qp_solvers):
    """Benchmark single gene deletion using MOMA."""
    model.solver = qp_solvers
    genes = ['b0008', 'b0114', 'b2276', 'b1779']
    benchmark(single_gene_deletion, model=model, gene_list=genes,
              method="moma", processes=1)


def test_single_gene_deletion_moma(model, qp_solvers):
    """Test single gene deletion using MOMA."""
    model.solver = qp_solvers
    # expected knockouts for textbook model
    growth_dict = {"b0008": 0.87, "b0114": 0.71, "b0116": 0.56,
                   "b2276": 0.11, "b1779": 0.00}

    result = single_gene_deletion(
        model=model,
        gene_list=list(growth_dict),
        method="moma",
        processes=1
    )["growth"]
    for gene, value in iteritems(growth_dict):
        assert np.isclose(result[frozenset([gene])], value,
                          atol=1E-02)


def test_single_gene_deletion_moma_reference(model, qp_solvers):
    """Test single gene deletion using MOMA (reference solution)."""
    model.solver = qp_solvers
    # expected knockouts for textbook model
    growth_dict = {"b0008": 0.87, "b0114": 0.71, "b0116": 0.56,
                   "b2276": 0.11, "b1779": 0.00}

    sol = model.optimize()
    result = single_gene_deletion(
        model=model,
        gene_list=list(growth_dict),
        method="moma",
        solution=sol,
        processes=1
    )["growth"]
    for gene, value in iteritems(growth_dict):
        assert np.isclose(result[frozenset([gene])], value,
                          atol=1E-02)


# Single gene deletion linear MOMA
def test_single_gene_deletion_linear_moma_benchmark(
        model, benchmark, all_solvers):
    """Benchmark single gene deletion using linear MOMA."""
    model.solver = all_solvers
    genes = ['b0008', 'b0114', 'b2276', 'b1779']
    benchmark(single_gene_deletion, model=model, gene_list=genes,
              method="linear moma", processes=1)


def test_single_gene_deletion_linear_moma(model, all_solvers):
    """Test single gene deletion using linear MOMA (reference solution)."""
    model.solver = all_solvers
    # expected knockouts for textbook model
    growth_dict = {"b0008": 0.87, "b0114": 0.76, "b0116": 0.65,
                   "b2276": 0.08, "b1779": 0.00}

    sol = model.optimize()
    result = single_gene_deletion(
        model=model,
        gene_list=list(growth_dict),
        method="linear moma",
        solution=sol,
        processes=1
    )["growth"]
    for gene, value in iteritems(growth_dict):
        assert np.isclose(result[frozenset([gene])], value,
                          atol=1E-02)


# Single gene deletion ROOM
def test_single_gene_deletion_room_benchmark(model, benchmark,
                                             all_solvers):
    """Benchmark single gene deletion using ROOM."""
    if all_solvers == "glpk":
        pytest.skip("GLPK is too slow to run ROOM.")
    model.solver = all_solvers
    genes = ['b0008', 'b0114', 'b2276', 'b1779']
    benchmark(single_gene_deletion, model=model, gene_list=genes,
              method="room", processes=1)


# Single gene deletion linear ROOM
def test_single_gene_deletion_linear_room_benchmark(
        model, benchmark, all_solvers):
    """Benchmark single gene deletion using linear ROOM."""
    model.solver = all_solvers
    genes = ['b0008', 'b0114', 'b2276', 'b1779']
    benchmark(single_gene_deletion, model=model, gene_list=genes,
              method="linear room", processes=1)


# Single reaction deletion
def test_single_reaction_deletion_benchmark(model, benchmark,
                                            all_solvers):
    """Benchmark single reaction deletion."""
    model.solver = all_solvers
    benchmark(single_reaction_deletion, model=model, processes=1)


def test_single_reaction_deletion(model, all_solvers):
    """Test single reaction deletion."""
    model.solver = all_solvers
    expected_results = {'FBA': 0.70404, 'FBP': 0.87392, 'CS': 0,
                        'FUM': 0.81430, 'GAPD': 0, 'GLUDy': 0.85139}
    result = single_reaction_deletion(
        model=model,
        reaction_list=list(expected_results),
        processes=1
    )['growth']
    for reaction, value in iteritems(expected_results):
        assert np.isclose(result[frozenset([reaction])], value,
                          atol=1E-05)


# Single reaction deletion ROOM
def test_single_reaction_deletion_room(room_model, room_solution,
                                       all_solvers):
    """Test single reaction deletion using ROOM."""
    room_model.solver = all_solvers
    expected = Series({'v1': 10.0, 'v2': 5.0, 'v3': 0.0, 'v4': 5.0,
                       'v5': 5.0, 'v6': 0.0, 'b1': 10.0, 'b2': 5.0,
                       'b3': 5.0}, index=['v1', 'v2', 'v3', 'v4',
                                          'v5', 'v6', 'b1', 'b2',
                                          'b3'])
    with room_model:
        room_model.reactions.v6.knock_out()
        add_room(room_model, solution=room_solution, delta=0.0, epsilon=0.0)
        room_sol = room_model.optimize()

    assert np.allclose(room_sol.fluxes, expected)


# Single reaction deletion linear ROOM
def test_single_reaction_deletion_linear_room(room_model, room_solution,
                                              all_solvers):
    """Test single reaction deletion using linear ROOM."""
    room_model.solver = all_solvers
    expected = Series({'v1': 10.0, 'v2': 5.0, 'v3': 0.0, 'v4': 5.0,
                       'v5': 5.0, 'v6': 0.0, 'b1': 10.0, 'b2': 5.0,
                       'b3': 5.0}, index=['v1', 'v2', 'v3', 'v4',
                                          'v5', 'v6', 'b1', 'b2',
                                          'b3'])
    with room_model:
        room_model.reactions.v6.knock_out()
        add_room(room_model, solution=room_solution, delta=0.0, epsilon=0.0,
                 linear=True)
        linear_room_sol = room_model.optimize()

    assert np.allclose(linear_room_sol.fluxes, expected)


# Double gene deletion
def test_double_gene_deletion_benchmark(large_model, benchmark):
    """Benchmark double gene deletion."""
    genes = ["b0726", "b4025", "b0724", "b0720", "b2935", "b2935",
             "b1276",
             "b1241"]
    benchmark(double_gene_deletion, large_model, gene_list1=genes,
              processes=1)


def test_double_gene_deletion(model):
    """Test double gene deletion."""
    genes = ["b0726", "b4025", "b0724", "b0720", "b2935", "b2935",
             "b1276",
             "b1241"]
    growth_dict = {'b0720': {'b0720': 0.0,
                             'b0724': 0.0,
                             'b0726': 0.0,
                             'b1241': 0.0,
                             'b1276': 0.0,
                             'b2935': 0.0,
                             'b4025': 0.0},
                   'b0724': {'b0720': 0.0,
                             'b0724': 0.814,
                             'b0726': 0.814,
                             'b1241': 0.814,
                             'b1276': 0.814,
                             'b2935': 0.814,
                             'b4025': 0.739},
                   'b0726': {'b0720': 0.0,
                             'b0724': 0.814,
                             'b0726': 0.858,
                             'b1241': 0.858,
                             'b1276': 0.858,
                             'b2935': 0.858,
                             'b4025': 0.857},
                   'b1241': {'b0720': 0.0,
                             'b0724': 0.814,
                             'b0726': 0.858,
                             'b1241': 0.874,
                             'b1276': 0.874,
                             'b2935': 0.874,
                             'b4025': 0.863},
                   'b1276': {'b0720': 0.0,
                             'b0724': 0.814,
                             'b0726': 0.858,
                             'b1241': 0.874,
                             'b1276': 0.874,
                             'b2935': 0.874,
                             'b4025': 0.863},
                   'b2935': {'b0720': 0.0,
                             'b0724': 0.814,
                             'b0726': 0.858,
                             'b1241': 0.874,
                             'b1276': 0.874,
                             'b2935': 0.874,
                             'b4025': 0.863},
                   'b4025': {'b0720': 0.0,
                             'b0724': 0.739,
                             'b0726': 0.857,
                             'b1241': 0.863,
                             'b1276': 0.863,
                             'b2935': 0.863,
                             'b4025': 0.863}}
    solution = double_gene_deletion(
        model, gene_list1=genes, processes=3)['growth']
    solution_one_process = double_gene_deletion(
        model, gene_list1=genes, processes=1)['growth']
    for (rxn_a, sub) in iteritems(growth_dict):
        for rxn_b, growth in iteritems(sub):
            sol = solution[frozenset([rxn_a, rxn_b])]
            sol_one = solution_one_process[frozenset([rxn_a, rxn_b])]
            assert round(sol, 3) == growth
            assert round(sol_one, 3) == growth


# Double reaction deletion
def test_double_reaction_deletion_benchmark(large_model, benchmark):
    """Benchmark double reaction deletion."""
    reactions = large_model.reactions[1::100]
    benchmark(double_reaction_deletion, large_model,
              reaction_list1=reactions)


def test_double_reaction_deletion(model):
    """Test double reaction deletion."""
    reactions = ['FBA', 'ATPS4r', 'ENO', 'FRUpts2']
    growth_dict = {
        "FBA": {
            "ATPS4r": 0.135,
            "ENO": float('nan'),
            "FRUpts2": 0.704
        },
        "ATPS4r": {
            "ENO": float('nan'),
            "FRUpts2": 0.374
        },
        "ENO": {
            "FRUpts2": 0.0
        },
    }

    solution = double_reaction_deletion(
        model, reaction_list1=reactions, processes=3)['growth']
    solution_one_process = double_reaction_deletion(
        model, reaction_list1=reactions, processes=1)['growth']
    for (rxn_a, sub) in iteritems(growth_dict):
        for rxn_b, growth in iteritems(sub):
            sol = solution[frozenset([rxn_a, rxn_b])]
            sol_one = solution_one_process[frozenset([rxn_a, rxn_b])]
            if math.isnan(growth):
                assert math.isnan(sol)
                assert math.isnan(sol_one)
            else:
                assert round(sol, 3) == growth
                assert round(sol_one, 3) == growth



# -*- coding: utf-8 -*-

"""Contains functions and helpers to obtain minimal growth media."""

import logging

import pandas as pd
from optlang.interface import OPTIMAL
from optlang.symbolics import Zero

from cobra.medium.boundary_types import find_boundary_types


LOGGER = logging.getLogger(__name__)


def add_linear_obj(model):
    """Add a linear version of a minimal medium to the model solver.

    Changes the optimization objective to finding the growth medium requiring
    the smallest total import flux::

        minimize sum |r_i| for r_i in import_reactions

    Arguments
    ---------
    model : cobra.Model
        The model to modify.
    """
    coefs = {}
    for rxn in find_boundary_types(model, "exchange"):
        export = len(rxn.reactants) == 1
        if export:
            coefs[rxn.reverse_variable] = 1
        else:
            coefs[rxn.forward_variable] = 1
    model.objective.set_linear_coefficients(coefs)
    model.objective.direction = "min"


def add_mip_obj(model):
    """Add a mixed-integer version of a minimal medium to the model.

    Changes the optimization objective to finding the medium with the least
    components::

        minimize size(R) where R part of import_reactions

    Arguments
    ---------
    model : cobra.model
        The model to modify.
    """
    if len(model.variables) > 1e4:
        LOGGER.warning("the MIP version of minimal media is extremely slow for"
                       " models that large :(")
    exchange_rxns = find_boundary_types(model, "exchange")
    big_m = max(abs(b) for r in exchange_rxns for b in r.bounds)
    prob = model.problem
    coefs = {}
    to_add = []
    for rxn in exchange_rxns:
        export = len(rxn.reactants) == 1
        indicator = prob.Variable("ind_" + rxn.id, lb=0, ub=1, type="binary")
        if export:
            vrv = rxn.reverse_variable
            indicator_const = prob.Constraint(
                vrv - indicator * big_m, ub=0, name="ind_constraint_" + rxn.id)
        else:
            vfw = rxn.forward_variable
            indicator_const = prob.Constraint(
                vfw - indicator * big_m, ub=0, name="ind_constraint_" + rxn.id)
        to_add.extend([indicator, indicator_const])
        coefs[indicator] = 1
    model.add_cons_vars(to_add)
    model.solver.update()
    model.objective.set_linear_coefficients(coefs)
    model.objective.direction = "min"


def _as_medium(exchanges, tolerance=1e-6, exports=False):
    """Convert a solution to medium.

    Arguments
    ---------
    exchanges : list of cobra.reaction
        The exchange reactions to consider.
    tolerance : positive double
        The absolute tolerance for fluxes. Fluxes with an absolute value
        smaller than this number will be ignored.
    exports : bool
        Whether to return export fluxes as well.

    Returns
    -------
    pandas.Series
        The "medium", meaning all active import fluxes in the solution.
    """
    LOGGER.debug("Formatting medium.")
    medium = pd.Series()
    for rxn in exchanges:
        export = len(rxn.reactants) == 1
        flux = rxn.flux
        if abs(flux) < tolerance:
            continue
        if export:
            medium[rxn.id] = -flux
        elif not export:
            medium[rxn.id] = flux
    if not exports:
        medium = medium[medium > 0]

    return medium


def minimal_medium(model, min_objective_value=0.1, exports=False,
                   minimize_components=False, open_exchanges=False):
    """
    Find the minimal growth medium for the model.

    Finds the minimal growth medium for the model which allows for
    model as well as individual growth. Here, a minimal medium can either
    be the medium requiring the smallest total import flux or the medium
    requiring the least components (ergo ingredients), which will be much
    slower due to being a mixed integer problem (MIP).

    Arguments
    ---------
    model : cobra.model
        The model to modify.
    min_objective_value : positive float or array-like object
        The minimum growth rate (objective) that has to be achieved.
    exports : boolean
        Whether to include export fluxes in the returned medium. Defaults to
        False which will only return import fluxes.
    minimize_components : boolean or positive int
        Whether to minimize the number of components instead of the total
        import flux. Might be more intuitive if set to True but may also be
        slow to calculate for large communities. If set to a number `n` will
        return up to `n` alternative solutions all with the same number of
        components.
    open_exchanges : boolean or number
        Whether to ignore currently set bounds and make all exchange reactions
        in the model possible. If set to a number all exchange reactions will
        be opened with (-number, number) as bounds.

    Returns
    -------
    pandas.Series, pandas.DataFrame or None
        A series giving the import flux for each required import
        reaction and (optionally) the associated export fluxes. All exchange
        fluxes are oriented into the import reaction e.g. positive fluxes
        denote imports and negative fluxes exports. If `minimize_components`
        is a number larger 1 may return a DataFrame where each column is a
        minimal medium. Returns None if the minimization is infeasible
        (for instance if min_growth > maximum growth rate).

    Notes
    -----
    Due to numerical issues the `minimize_components` option will usually only
    minimize the number of "large" import fluxes. Specifically, the detection
    limit is given by ``integrality_tolerance * max_bound`` where ``max_bound``
    is the largest bound on an import reaction. Thus, if you are interested
    in small import fluxes as well you may have to adjust the integrality
    tolerance at first with
    `model.solver.configuration.tolerances.integrality = 1e-7` for instance.
    However, this will be *very* slow for large models especially with GLPK.

    """
    exchange_rxns = find_boundary_types(model, "exchange")
    if isinstance(open_exchanges, bool):
        open_bound = 1000
    else:
        open_bound = open_exchanges

    with model as mod:
        if open_exchanges:
            LOGGER.debug("Opening exchanges for %d imports.",
                         len(exchange_rxns))
            for rxn in exchange_rxns:
                rxn.bounds = (-open_bound, open_bound)
        LOGGER.debug("Applying objective value constraints.")
        obj_const = mod.problem.Constraint(
            mod.objective.expression, lb=min_objective_value,
            name="medium_obj_constraint")
        mod.add_cons_vars([obj_const])
        mod.solver.update()
        mod.objective = Zero
        LOGGER.debug("Adding new media objective.")
        tol = mod.solver.configuration.tolerances.feasibility

        if minimize_components:
            add_mip_obj(mod)
            if isinstance(minimize_components, bool):
                minimize_components = 1
            seen = set()
            best = num_components = mod.slim_optimize()
            if mod.solver.status != OPTIMAL:
                LOGGER.warning("Minimization of medium was infeasible.")
                return None
            exclusion = mod.problem.Constraint(Zero, ub=0)
            mod.add_cons_vars([exclusion])
            mod.solver.update()
            media = []
            for i in range(minimize_components):
                LOGGER.info("Finding alternative medium #%d.", (i + 1))
                vars = [mod.variables["ind_" + s] for s in seen]
                if len(seen) > 0:
                    exclusion.set_linear_coefficients(
                        dict.fromkeys(vars, 1))
                    exclusion.ub = best - 1
                num_components = mod.slim_optimize()
                if mod.solver.status != OPTIMAL or num_components > best:
                    break
                medium = _as_medium(exchange_rxns, tol, exports=exports)
                media.append(medium)
                seen.update(medium[medium > 0].index)
            if len(media) > 1:
                medium = pd.concat(media, axis=1, sort=True).fillna(0.0)
                medium.sort_index(axis=1, inplace=True)
            else:
                medium = media[0]
        else:
            add_linear_obj(mod)
            mod.slim_optimize()
            if mod.solver.status != OPTIMAL:
                LOGGER.warning("Minimization of medium was infeasible.")
                return None
            medium = _as_medium(exchange_rxns, tol, exports=exports)

    return medium



# -*- coding: utf-8 -*-

"""Test functionalities provided by sbml.py"""

from __future__ import absolute_import

from io import BytesIO
from os.path import getsize, join

import pytest

from cobra import io
from cobra.test.test_io.conftest import compare_models


try:
    import libsbml
except ImportError:
    libsbml = None


@pytest.fixture(scope="function")
def mini_fbc1_model(data_directory):
    return io.read_legacy_sbml(join(data_directory, "mini_fbc1.xml"))


@pytest.fixture(scope="function")
def mini_cobra_model(data_directory):
    return io.read_legacy_sbml(join(data_directory, "mini_cobra.xml"))


# TODO: parametrize the arguments after pytest.fixture_request()
# is approved
@pytest.mark.skipif(libsbml is None, reason="libsbml unavailable.")
def test_read_sbml_model(data_directory, mini_fbc1_model, mini_cobra_model):
    """Test the reading of a model from SBML v2."""
    mini_fbc1 = io.read_legacy_sbml(join(data_directory, "mini_fbc1.xml"))
    mini_cobra = io.read_legacy_sbml(join(data_directory, "mini_cobra.xml"))
    assert compare_models(mini_fbc1_model, mini_fbc1) is None
    assert compare_models(mini_cobra_model, mini_cobra) is None


@pytest.mark.skipif(libsbml is None, reason="libsbml unavailable.")
def test_read_file_handle(data_directory, mini_model):
    """Test the reading of a model passed as a file handle."""
    with open(join(data_directory, "mini_cobra.xml"), "rb") as file_:
        model_stream = BytesIO(file_.read())
    read_model = io.read_sbml_model(model_stream)
    assert compare_models(mini_model, read_model) is None


# TODO: parametrize the arguments after pytest.fixture_request()
# is approved
@pytest.mark.skipif(libsbml is None, reason="libsbml unavailable.")
def test_write_sbml_model(tmpdir, mini_fbc1_model, mini_cobra_model):
    """Test the writing of a model to SBML v2."""
    mini_fbc1_output_file = tmpdir.join("mini_fbc1.xml")
    mini_cobra_output_file = tmpdir.join("mini_cobra.xml")

    # convert to str object before passing the filename
    io.write_legacy_sbml(mini_fbc1_model, str(mini_fbc1_output_file),
                         use_fbc_package=True)
    io.write_legacy_sbml(mini_cobra_model, str(mini_cobra_output_file),
                         use_fbc_package=False)

    assert mini_fbc1_output_file.check()
    assert mini_cobra_output_file.check()



# -*- coding: utf-8 -*-

"""Test functionalities of Flux Variability Analysis."""

from __future__ import absolute_import

import numpy as np
import pytest
from six import iteritems

from cobra.exceptions import Infeasible
from cobra.flux_analysis.variability import (
    find_blocked_reactions, find_essential_genes, find_essential_reactions,
    flux_variability_analysis)


# FVA
def test_flux_variability_benchmark(large_model, benchmark,
                                    all_solvers):
    """Benchmark FVA."""
    large_model.solver = all_solvers
    benchmark(flux_variability_analysis, large_model,
              reaction_list=large_model.reactions[1::3],
              processes=1)


def test_pfba_flux_variability(model, pfba_fva_results,
                               fva_results, all_solvers):
    """Test FVA using pFBA."""
    model.solver = all_solvers
    with pytest.warns(UserWarning):
        flux_variability_analysis(
            model, pfba_factor=0.1, reaction_list=model.reactions[1::3],
            processes=1)
    fva_out = flux_variability_analysis(
        model, pfba_factor=1.1, reaction_list=model.reactions,
        processes=1)
    fva_out.sort_index(inplace=True)
    assert np.allclose(fva_out, pfba_fva_results)
    abs_fva_out = fva_out.dropna().abs()
    abs_fva_results = fva_results.dropna().abs()
    comparison = np.isclose(abs_fva_out, abs_fva_results) | (
            abs_fva_out < abs_fva_results)
    assert comparison["minimum"].all()
    assert comparison["maximum"].all()


def test_loopless_pfba_fva(model):
    loop_reactions = [model.reactions.get_by_id(rid)
                      for rid in ("FRD7", "SUCDi")]
    fva_loopless = flux_variability_analysis(
        model, pfba_factor=1.1, reaction_list=loop_reactions,
        loopless=True, processes=1)
    assert np.allclose(fva_loopless["maximum"],
                       fva_loopless["minimum"])


def test_flux_variability(model, fva_results, all_solvers):
    """Test FVA."""
    model.solver = all_solvers
    fva_out = flux_variability_analysis(model, reaction_list=model.reactions,
                                        processes=1)
    fva_out.sort_index(inplace=True)
    assert np.allclose(fva_out, fva_results)


def test_parallel_flux_variability(model, fva_results, all_solvers):
    """Test parallel FVA."""
    model.solver = all_solvers
    fva_out = flux_variability_analysis(model, processes=2)
    fva_out.sort_index(inplace=True)
    assert np.allclose(fva_out, fva_results)


# Loopless FVA
def test_flux_variability_loopless_benchmark(model, benchmark,
                                             all_solvers):
    """Benchmark loopless FVA."""
    model.solver = all_solvers
    benchmark(flux_variability_analysis, model, loopless=True,
              reaction_list=model.reactions[1::3])


def test_flux_variability_loopless(model, all_solvers):
    """Test loopless FVA."""
    model.solver = all_solvers
    loop_reactions = [model.reactions.get_by_id(rid)
                      for rid in ("FRD7", "SUCDi")]
    fva_normal = flux_variability_analysis(
        model, reaction_list=loop_reactions)
    fva_loopless = flux_variability_analysis(
        model, reaction_list=loop_reactions, loopless=True)

    assert not np.allclose(fva_normal["maximum"],
                           fva_normal["minimum"])
    assert np.allclose(fva_loopless["maximum"],
                       fva_loopless["minimum"])


# Internals (essentiality, blocked reactions)
def test_fva_data_frame(model):
    """Test DataFrame obtained from FVA."""
    df = flux_variability_analysis(model)
    assert np.all([df.columns.values == ['minimum', 'maximum']])


def test_fva_infeasible(model):
    """Test FVA infeasibility."""
    infeasible_model = model.copy()
    infeasible_model.reactions.get_by_id("EX_glc__D_e").lower_bound = 0
    # ensure that an infeasible model does not run FVA
    with pytest.raises(Infeasible):
        flux_variability_analysis(infeasible_model)


def test_fva_minimization(model):
    """Test minimization using FVA."""
    model.objective = model.reactions.EX_glc__D_e
    model.objective_direction = 'min'
    solution = flux_variability_analysis(model, fraction_of_optimum=.95)
    assert solution.at['EX_glc__D_e', 'minimum'] == -10.0
    assert solution.at['EX_glc__D_e', 'maximum'] == -9.5


def test_find_blocked_reactions_solver_none(model):
    """Test find_blocked_reactions() [no specific solver]."""
    result = find_blocked_reactions(model, model.reactions[40:46])
    assert result == ['FRUpts2']


def test_essential_genes(model):
    """Test find_essential_genes()."""
    essential_genes = {'b2779', 'b1779', 'b0720', 'b2416',
                       'b2926', 'b1136', 'b2415'}
    observed_essential_genes = {g.id for g in
                                find_essential_genes(model)}
    assert observed_essential_genes == essential_genes


def test_essential_reactions(model):
    """Test find_blocked_reactions()."""
    essential_reactions = {'GLNS', 'Biomass_Ecoli_core', 'PIt2r',
                           'GAPD',
                           'ACONTb', 'EX_nh4_e', 'ENO', 'EX_h_e',
                           'EX_glc__D_e', 'ICDHyr', 'CS', 'NH4t',
                           'GLCpts',
                           'PGM', 'EX_pi_e', 'PGK', 'RPI', 'ACONTa'}
    observed_essential_reactions = {r.id for r in
                                    find_essential_reactions(model)}
    assert observed_essential_reactions == essential_reactions


def test_find_blocked_reactions(model, all_solvers):
    """Test find_blocked_reactions()."""
    model.solver = all_solvers
    result = find_blocked_reactions(model, model.reactions[40:46])
    assert result == ['FRUpts2']

    result = find_blocked_reactions(model, model.reactions[42:48])
    assert set(result) == {'FUMt2_2', 'FRUpts2'}

    result = find_blocked_reactions(model, model.reactions[30:50],
                                    open_exchanges=True)
    assert result == []



import json
import pandas as pd
import argparse
import re
from os.path import basename

pd.set_option('display.width', 200)


def benchmark_to_df(json_file):
    with open(json_file) as jf:
        content = json.load(jf)
        df = pd.DataFrame(columns=("test", "time [ms] "))
        for b in content["benchmarks"]:
            df = df.append({"test": b["name"],
                            "time [ms] ": b["stats"]["mean"] * 1000.0},
                           ignore_index=True)
        return df


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="""
    compare cobrapy benchmarks.
    Run pytest with
    pytest --benchmark-save=without-cache --benchmark-min-rounds=20
    then compare saved json files with this script.
    """)
    parser.add_argument('first', help='first json file')
    parser.add_argument('second', help='second json file')
    args = parser.parse_args()

    first = benchmark_to_df(args.first)
    second = benchmark_to_df(args.second)
    re_name = '^[0-9]+_(.+).json$'
    first_name = re.findall(re_name, basename(args.first))[0]
    second_name = re.findall(re_name, basename(args.second))[0]
    both = pd.merge(first, second, how="inner", on="test",
                    suffixes=(first_name, second_name))
    both["fraction"] = both.iloc[:, 2] / both.iloc[:, 1]
    print(both.sort_values(by="fraction"))




# -*- coding: utf-8 -*-

"""Test functionalities of pFBA."""

import warnings

import numpy as np
import pytest

from cobra.exceptions import Infeasible
from cobra.flux_analysis.parsimonious import add_pfba, pfba


def test_pfba_benchmark(large_model, benchmark, all_solvers):
    """Benchmark pFBA functionality."""
    large_model.solver = all_solvers
    benchmark(pfba, large_model)


def test_pfba(model, all_solvers):
    """Test pFBA functionality."""
    model.solver = all_solvers
    with model:
        add_pfba(model)
        with pytest.raises(ValueError):
            add_pfba(model)

    expression = model.objective.expression
    n_constraints = len(model.constraints)
    solution = pfba(model)
    assert solution.status == "optimal"
    assert solution.fluxes["Biomass_Ecoli_core"] == \
        pytest.approx(0.8739, abs=1e-4, rel=0.0)
    assert solution.fluxes.abs().sum() == \
        pytest.approx(518.4221, abs=1e-4, rel=0.0)

    # test changes to model reverted
    assert expression == model.objective.expression
    assert len(model.constraints) == n_constraints

    # needed?
    # Test desired_objective_value
    # desired_objective = 0.8
    # pfba(model, solver=solver,
    #                       desired_objective_value=desired_objective)
    # abs_x = [abs(i) for i in model.solution.x]
    # assert model.solution.status == "optimal"
    # assert abs(model.solution.f - desired_objective) < 0.001
    # assert abs(sum(abs_x) - 476.1594) < 0.001

    # TODO: parametrize fraction (DRY it up)
    # Test fraction_of_optimum
    solution = pfba(model, fraction_of_optimum=0.95)
    assert solution.status == "optimal"
    assert solution.fluxes["Biomass_Ecoli_core"] == pytest.approx(
        0.95 * 0.8739, abs=1e-4, rel=0.0)
    abs_x = [abs(i) for i in solution.fluxes.values]
    assert sum(abs_x) == pytest.approx(493.4400, abs=1e-4, rel=0.0)

    # Infeasible solution
    model.reactions.ATPM.lower_bound = 500
    with warnings.catch_warnings():
        warnings.simplefilter("error", UserWarning)
        with pytest.raises((UserWarning, Infeasible, ValueError)):
            pfba(model)



# -*- coding: utf-8 -*-

"""Test functions of metabolite.py"""

import pytest

from cobra.core import Metabolite


def test_metabolite_formula():
    met = Metabolite("water")
    met.formula = "H2O"
    assert met.elements == {"H": 2, "O": 1}
    assert met.formula_weight == 18.01528


def test_formula_element_setting(model):
    met = model.metabolites[1]
    orig_formula = str(met.formula)
    orig_elements = dict(met.elements)
    met.formula = ''
    assert met.elements == {}
    met.elements = orig_elements
    assert met.formula == orig_formula


def test_set_id(solved_model):
    solution, model = solved_model
    met = Metabolite("test")
    with pytest.raises(TypeError):
        setattr(met, 'id', 1)
    model.add_metabolites([met])
    with pytest.raises(ValueError):
        setattr(met, "id", 'g6p_c')
    met.id = "test2"
    assert "test2" in model.metabolites
    assert "test" not in model.metabolites


def test_remove_from_model(solved_model):
    solution, model = solved_model
    met = model.metabolites.get_by_id("g6p_c")
    met.remove_from_model()
    assert not (met.id in model.metabolites)
    assert not (met.id in model.constraints)


def test_repr_html_(model):
    assert '<table>' in model.metabolites.h2o_c._repr_html_()



# -*- coding: utf-8 -*-

"""Lists and annotations for compartment names and reactions.

Please send a PR if you want to add something here :)
"""

excludes = {
    "demand": ["SN_", "SK_", "sink", "EX_", "exchange"],
    "exchange": ["demand", "DM_", "biosynthesis", "transcription",
                 "replication", "SN_", "SK_", "sink"],
    "sink": ["demand", "DM_", "biosynthesis", "transcription",
             "replication", "EX_", "exchange"]
}
"""A list of sub-strings in reaction IDs that usually indicate
that the reaction is *not* a reaction of the specified type."""

sbo_terms = {"demand": "SBO:0000628",
             "exchange": "SBO:0000627",
             "sink": "SBO:0000632",
             "biomass": "SBO:0000629",
             "pseudoreaction": "SBO:0000631"}
"""SBO term identifiers for various boundary types."""

compartment_shortlist = {
    "ce": ["cell envelope"],
    "c": ["cytoplasm", "cytosol", "default", "in", "intra cellular",
          "intracellular", "intracellular region", "intracellular space"],
    "er": ["endoplasmic reticulum"],
    "erm": ["endoplasmic reticulum membrane"],
    "e": ["extracellular", "extraorganism", "out", "extracellular space",
          "extra organism", "extra cellular", "extra-organism", "external",
          "external medium"],
    "f": ["flagellum", "bacterial-type flagellum"],
    "g": ["golgi", "golgi apparatus"],
    "gm": ["golgi membrane"],
    "h": ["chloroplast"],
    "l": ["lysosome"],
    "im": ["mitochondrial intermembrane space"],
    "mm": ["mitochondrial membrane"],
    "m": ["mitochondrion", "mitochondria"],
    "n": ["nucleus"],
    "p": ["periplasm", "periplasmic space"],
    "x": ["peroxisome", "glyoxysome"],
    "u": ["thylakoid"],
    "vm": ["vacuolar membrane"],
    "v": ["vacuole"],
    "w": ["cell wall"],
    "s": ["eyespot", "eyespot apparatus", "stigma"]
}
"""A list of common compartment abbreviations and alternative names."""



# -*- coding: utf-8 -*-

"""Test functions of util.py"""

from __future__ import absolute_import

from cobra.util import show_versions


def test_show_versions(capsys):
    show_versions()
    captured = capsys.readouterr()
    lines = captured.out.split("\n")
    assert lines[1].startswith("System Information")
    assert lines[2].startswith("==================")
    assert lines[3].startswith("OS")
    assert lines[4].startswith("OS-release")
    assert lines[5].startswith("Python")

    assert lines[7].startswith("Package Versions")
    assert lines[8].startswith("================")



# -*- coding: utf-8 -*-

"""Test functionalities of gapfilling."""

from __future__ import absolute_import

from cobra.core import Metabolite, Model, Reaction
from cobra.flux_analysis.gapfilling import GapFiller, gapfill


def test_gapfilling(salmonella):
    """Test Gapfilling."""
    m = Model()
    m.add_metabolites([Metabolite(m_id) for m_id in ["a", "b", "c"]])
    exa = Reaction("EX_a")
    exa.add_metabolites({m.metabolites.a: 1})
    b2c = Reaction("b2c")
    b2c.add_metabolites({m.metabolites.b: -1, m.metabolites.c: 1})
    dmc = Reaction("DM_c")
    dmc.add_metabolites({m.metabolites.c: -1})
    m.add_reactions([exa, b2c, dmc])
    m.objective = 'DM_c'

    universal = Model()
    a2b = Reaction("a2b")
    a2d = Reaction("a2d")
    universal.add_reactions([a2b, a2d])
    a2b.build_reaction_from_string("a --> b", verbose=False)
    a2d.build_reaction_from_string("a --> d", verbose=False)

    # # GrowMatch
    # result = gapfilling.growMatch(m, universal)[0]
    result = gapfill(m, universal)[0]
    assert len(result) == 1
    assert result[0].id == "a2b"

    # # SMILEY
    # result = gapfilling.SMILEY(m, "b", universal)[0]
    with m:
        m.objective = m.add_boundary(m.metabolites.b, type='demand')
        result = gapfill(m, universal)[0]
        assert len(result) == 1
        assert result[0].id == "a2b"

    # # 2 rounds of GrowMatch with exchange reactions
    # result = gapfilling.growMatch(m, None, ex_rxns=True, iterations=2)
    result = gapfill(m, None, exchange_reactions=True,
                     iterations=2)
    assert len(result) == 2
    assert len(result[0]) == 1
    assert len(result[1]) == 1
    assert {i[0].id for i in result} == {"EX_b", "EX_c"}

    # # Gapfilling solution adds metabolites not present in original model
    # test for when demand = T
    # a demand reaction must be added to clear new metabolite
    universal_noDM = Model()
    a2b = Reaction("a2b")
    universal_noDM.add_reactions([a2b])
    a2b.build_reaction_from_string("a --> b + d", verbose=False)
    result = gapfill(m, universal_noDM,
                     exchange_reactions=False,
                     demand_reactions=True)[0]
    # add reaction a2b and demand reaction to clear met d
    assert len(result) == 2
    assert "a2b" in [x.id for x in result]

    # test for when demand = False
    # test for when metabolites are added to the model and
    # must be cleared by other reactions in universal model
    # (i.e. not necessarily a demand reaction)
    universal_withDM = universal_noDM.copy()
    d_dm = Reaction("d_dm")
    universal_withDM.add_reactions([d_dm])
    d_dm.build_reaction_from_string("d -->", verbose=False)
    result = gapfill(m, universal_withDM,
                     exchange_reactions=False,
                     demand_reactions=False)[0]
    assert len(result) == 2
    assert "a2b" in [x.id for x in result]

    # somewhat bigger model
    universal = Model("universal_reactions")
    with salmonella as model:
        for i in [i.id for i in model.metabolites.f6p_c.reactions]:
            reaction = model.reactions.get_by_id(i)
            universal.add_reactions([reaction.copy()])
            model.remove_reactions([reaction])
        gf = GapFiller(model, universal,
                       penalties={'TKT2': 1e3},
                       demand_reactions=False)
        solution = gf.fill()
        assert 'TKT2' not in {r.id for r in solution[0]}
        assert gf.validate(solution[0])



# -*- coding: utf-8 -*-

"""Test functionalities of summary methods."""

from __future__ import absolute_import

import re
import sys
from contextlib import contextmanager

import pytest
from six import StringIO

from cobra.flux_analysis.parsimonious import pfba


@contextmanager
def captured_output():
    """A context manager to test the IO summary methods."""
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr

    finally:
        sys.stdout, sys.stderr = old_out, old_err


def check_line(output, expected_entries,
               pattern=re.compile(r"\s")):
    """Ensure each expected entry is in the output."""
    output_set = set(
        pattern.sub("", line) for line in output.splitlines())
    for elem in expected_entries:
        assert pattern.sub("", elem) in output_set


def check_in_line(output, expected_entries,
                  pattern=re.compile(r"\s")):
    """Ensure each expected entry is contained in the output."""
    output_strip = [pattern.sub("", line) for line in
                    output.splitlines()]
    for elem in expected_entries:
        assert any(
            pattern.sub("", elem) in line for line in output_strip), \
            "Not found: {} in:\n{}".format(pattern.sub("", elem),
                                           "\n".join(output_strip))


@pytest.mark.parametrize("names", [False, True])
def test_model_summary_previous_solution(model, opt_solver, names):
    """Test summary of previous solution."""
    model.solver = opt_solver
    solution = model.optimize()
    rxn_test = model.exchanges[0]
    if names:
        met_test = list(rxn_test.metabolites.keys())[0].name
    else:
        met_test = list(rxn_test.metabolites.keys())[0].id

    solution.fluxes[rxn_test.id] = 321

    with captured_output() as (out, _):
        model.summary(solution, names=names)
    check_in_line(out.getvalue(), [met_test + '321'])


@pytest.mark.parametrize("names", [False, True])
def test_model_summary(model, opt_solver, names):
    """Test model summary."""
    model.solver = opt_solver
    # test non-fva version (these should be fixed for textbook model)
    if names:
        expected_entries = [
            'O2      21.8',
            'D-Glucose  10',
            'Ammonium      4.77',
            'Phosphate       3.21',
            'H2O  29.2',
            'CO2  22.8',
            'H+    17.5',
            'Biomass_Ecol...  0.874',
        ]
    else:
        expected_entries = [
            'o2_e      21.8',
            'glc__D_e  10',
            'nh4_e      4.77',
            'pi_e       3.21',
            'h2o_e  29.2',
            'co2_e  22.8',
            'h_e    17.5',
            'Biomass_Ecol...  0.874',
        ]
    # Need to use a different method here because
    # there are multiple entries per line.
    model.optimize()
    with captured_output() as (out, _):
        model.summary(names=names)
    check_in_line(out.getvalue(), expected_entries)

    # with model:
    #     model.objective = model.exchanges[0]
    #     model.summary()


@pytest.mark.parametrize("fraction", [0.95])
def test_model_summary_with_fva(model, opt_solver, fraction):
    """Test model summary (using FVA)."""
    if opt_solver == "optlang-gurobi":
        pytest.xfail("FVA currently buggy")
    # test non-fva version (these should be fixed for textbook model
    expected_entries = [
        'idFluxRangeidFluxRangeBiomass_Ecol...0.874',
        'o2_e       21.8   [19.9, 23.7]'
        'h2o_e       29.2  [25, 30.7]',
        'glc__D_e   10     [9.52, 10]'
        'co2_e       22.8  [18.9, 24.7]',
        'nh4_e       4.77  [4.53, 5.16]'
        'h_e         17.5  [16.7, 22.4]',
        'pi_e        3.21  [3.05, 3.21]'
        'for_e        0    [0, 5.72]',
        'ac_e         0    [0, 1.91]',
        'pyr_e        0    [0, 1.27]',
        'lac__D_e     0    [0, 1.07]',
        'succ_e       0    [0, 0.837]',
        'glu__L_e     0    [0, 0.636]',
        'akg_e        0    [0, 0.715]',
        'etoh_e       0    [0, 1.11]',
        'acald_e      0    [0, 1.27]',
    ]
    # Need to use a different method here because
    # there are multiple entries per line.
    model.solver = opt_solver
    solution = model.optimize()
    with captured_output() as (out, _):
        model.summary(solution, fva=fraction)
    check_in_line(out.getvalue(), expected_entries)


@pytest.mark.parametrize("met", ["q8_c"])
def test_metabolite_summary_previous_solution(
        model, opt_solver, met):
    """Test metabolite summary of previous solution."""
    model.solver = opt_solver
    solution = pfba(model)
    model.metabolites.get_by_id(met).summary(solution)


@pytest.mark.parametrize("met, names", [
    ("q8_c", False),
    ("q8_c", True)
])
def test_metabolite_summary(model, opt_solver, met, names):
    """Test metabolite summary."""
    model.solver = opt_solver
    model.optimize()
    with captured_output() as (out, _):
        model.metabolites.get_by_id(met).summary(names=names)

    if names:
        expected_entries = [
            'PRODUCING REACTIONS -- Ubiquinone-8 (q8_c)',
            '%       FLUX  RXN ID      REACTION',
            '100%   43.6   cytochr...  2.0 H+ + 0.5 O2 + Ubiquinol-8 --> '
            'H2O + 2.0 H+ ...',
            'CONSUMING REACTIONS -- Ubiquinone-8 (q8_c)',
            '%       FLUX  RXN ID      REACTION',
            '88%    38.5   NADH de...  4.0 H+ + Nicotinamide adenine '
            'dinucleotide - re...',
            '12%     5.06  succina...  Ubiquinone-8 + Succinate --> '
            'Fumarate + Ubiquin...'
        ]
    else:
        expected_entries = [
            'PRODUCING REACTIONS -- Ubiquinone-8 (q8_c)',
            '%       FLUX  RXN ID    REACTION',
            '100%   43.6   CYTBD     '
            '2.0 h_c + 0.5 o2_c + q8h2_c --> h2o_c + 2.0 h_e...',
            'CONSUMING REACTIONS -- Ubiquinone-8 (q8_c)',
            '%       FLUX  RXN ID    REACTION',
            '88%    38.5   NADH16    '
            '4.0 h_c + nadh_c + q8_c --> 3.0 h_e + nad_c + q...',
            '12%     5.06  SUCDi     q8_c + succ_c --> fum_c + q8h2_c',
        ]

    check_in_line(out.getvalue(), expected_entries)


@pytest.mark.parametrize("fraction, met", [(0.99, "fdp_c")])
def test_metabolite_summary_with_fva(model, opt_solver, fraction,
                                     met):
    """Test metabolite summary (using FVA)."""
    #     pytest.xfail("FVA currently buggy")

    model.solver = opt_solver
    model.optimize()
    with captured_output() as (out, _):
        model.metabolites.get_by_id(met).summary(fva=fraction)

    expected_entries = [
        'PRODUCING REACTIONS -- D-Fructose 1,6-bisphosphate (fdp_c)',
        '%       FLUX  RANGE         RXN ID    REACTION',
        '100%    7.48  [6.17, 9.26]  PFK       '
        'atp_c + f6p_c --> adp_c + fdp_c + h_c',
        'CONSUMING REACTIONS -- D-Fructose 1,6-bisphosphate (fdp_c)',
        '%       FLUX  RANGE         RXN ID    REACTION',
        '100%    7.48  [6.17, 8.92]  FBA       fdp_c <=> dhap_c + g3p_c',
        '0%      0     [0, 1.72]     FBP       '
        'fdp_c + h2o_c --> f6p_c + pi_c',
    ]

    check_line(out.getvalue(), expected_entries)



# -*- coding: utf-8 -*-

"""Test functionalities of loopless.py"""

from __future__ import absolute_import

import pytest
from optlang.interface import INFEASIBLE, OPTIMAL

import cobra.util.solver as sutil
from cobra.core import Metabolite, Model, Reaction
from cobra.flux_analysis.loopless import add_loopless, loopless_solution


def construct_ll_test_model():
    """Construct test model."""
    test_model = Model()
    test_model.add_metabolites(Metabolite("A"))
    test_model.add_metabolites(Metabolite("B"))
    test_model.add_metabolites(Metabolite("C"))
    EX_A = Reaction("EX_A")
    EX_A.add_metabolites({test_model.metabolites.A: 1})
    DM_C = Reaction("DM_C")
    DM_C.add_metabolites({test_model.metabolites.C: -1})
    v1 = Reaction("v1")
    v1.add_metabolites({test_model.metabolites.A: -1,
                        test_model.metabolites.B: 1})
    v2 = Reaction("v2")
    v2.add_metabolites({test_model.metabolites.B: -1,
                        test_model.metabolites.C: 1})
    v3 = Reaction("v3")
    v3.add_metabolites({test_model.metabolites.C: -1,
                        test_model.metabolites.A: 1})
    test_model.add_reactions([EX_A, DM_C, v1, v2, v3])
    DM_C.objective_coefficient = 1
    return test_model


@pytest.fixture(scope="function", params=[s for s in ["glpk", "cplex",
                                                      "gurobi"]
                                          if s in sutil.solvers])
def ll_test_model(request):
    """Return test model set with different solvers."""
    test_model = construct_ll_test_model()
    test_model.solver = request.param
    return test_model


def test_loopless_benchmark_before(benchmark):
    """Benchmark initial condition."""
    test_model = construct_ll_test_model()

    def _():
        with test_model:
            add_loopless(test_model)
            test_model.optimize()

    benchmark(_)


def test_loopless_benchmark_after(benchmark):
    """Benchmark final condition."""
    test_model = construct_ll_test_model()
    benchmark(loopless_solution, test_model)


def test_loopless_solution(ll_test_model):
    """Test loopless_solution()."""
    solution_feasible = loopless_solution(ll_test_model)
    ll_test_model.reactions.v3.lower_bound = 1
    ll_test_model.optimize()
    solution_infeasible = loopless_solution(ll_test_model)
    assert solution_feasible.fluxes["v3"] == 0.0
    assert solution_infeasible.fluxes["v3"] == 1.0


def test_loopless_solution_fluxes(model):
    """Test fluxes of loopless_solution()"""
    fluxes = model.optimize().fluxes
    ll_solution = loopless_solution(model, fluxes=fluxes)
    assert len(ll_solution.fluxes) == len(model.reactions)


def test_add_loopless(ll_test_model):
    """Test add_loopless()."""
    add_loopless(ll_test_model)
    feasible_status = ll_test_model.optimize().status
    ll_test_model.reactions.v3.lower_bound = 1
    ll_test_model.slim_optimize()
    infeasible_status = ll_test_model.solver.status
    assert feasible_status == OPTIMAL
    assert infeasible_status == INFEASIBLE



# -*- coding: utf-8 -*-
from matplotlib.pyplot import figure, xlim, ylim, gca, arrow, text, scatter
from mpl_toolkits.axes_grid.axislines import SubplotZero
from numpy import linspace, arange, sqrt, pi, sin, cos, sign
from IPython.display import set_matplotlib_formats

set_matplotlib_formats('png', 'pdf')


# axis style
def make_plot_ax():
    fig = figure(figsize=(6, 5))
    ax = SubplotZero(fig, 111)
    fig.add_subplot(ax)
    for direction in ["xzero", "yzero"]:
        ax.axis[direction].set_axisline_style("-|>")
        ax.axis[direction].set_visible(True)
    for direction in ["left", "right", "bottom", "top"]:
        ax.axis[direction].set_visible(False)
    xlim(-0.1, 2.1)
    ylim(xlim())
    ticks = [0.5 * i for i in range(1, 5)]
    labels = [str(i) if i == int(i) else "" for i in ticks]
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xticklabels(labels)
    ax.set_yticklabels(labels)
    ax.axis["yzero"].set_axis_direction("left")
    return ax


def plot_qp1():
    ax = make_plot_ax()
    ax.plot((0, 2), (2, 0), 'b')
    ax.plot([1], [1], 'bo')

    # circular grid
    for r in sqrt(2.) + 0.125 * arange(-11, 6):
        t = linspace(0., pi/2., 100)
        ax.plot(r * cos(t), r * sin(t), '-.', color="gray")


def plot_qp2():
    ax = make_plot_ax()
    ax.plot((0, 2), (2, 0), 'b')
    ax.plot([0.5], [1.5], 'bo')

    yrange = linspace(1, 2, 11)
    for r in [y ** 2 / 2. - y for y in yrange]:
        t = linspace(-sqrt(2 * r + 1) + 0.000001,
                     sqrt(2 * r + 1) - 0.000001, 1000)
        ax.plot(abs(t), [1 + sqrt(abs(2 * r + 1 - x ** 2)) *
                         sign(x) for x in t],
                '-.', color="gray")


def plot_loop():
    figure(figsize=(10.5, 4.5), frameon=False)
    gca().axis("off")
    xlim(0.5, 3.5)
    ylim(0.7, 2.2)
    arrow_params = {"head_length": 0.08, "head_width": 0.1, "ec": "k",
                    "fc": "k"}
    text_params = {"fontsize": 25, "horizontalalignment": "center",
                   "verticalalignment": "center"}
    arrow(0.5, 1, 0.85, 0, **arrow_params)  # EX_A
    arrow(1.5, 1, 0.425, 0.736, **arrow_params)  # v1
    arrow(2.04, 1.82, 0.42, -0.72, **arrow_params)  # v2
    arrow(2.4, 1, -0.75, 0, **arrow_params)  # v3
    arrow(2.6, 1, 0.75, 0, **arrow_params)
    # reaction labels
    text(0.9, 1.15, "EX_A", **text_params)
    text(1.6, 1.5, r"v$_1$", **text_params)
    text(2.4, 1.5, r"v$_2$", **text_params)
    text(2, 0.85, r"v$_3$", **text_params)
    text(2.9, 1.15, "DM_C", **text_params)
    # metabolite labels
    scatter(1.5, 1, s=250, color='#c994c7')
    text(1.5, 0.9, "A", **text_params)
    scatter(2, 1.84, s=250, color='#c994c7')
    text(2, 1.95, "B", **text_params)
    scatter(2.5, 1, s=250, color='#c994c7')
    text(2.5, 0.9, "C", **text_params)



# -*- coding: utf-8 -*-

"""Test functions of reaction.py"""

from __future__ import absolute_import

import warnings

import numpy as np
import pytest
import six

from cobra.core import Metabolite, Model, Reaction


stable_optlang = ["glpk", "cplex", "gurobi"]


def test_gpr():
    model = Model()
    reaction = Reaction("test")

    # Set GPR to a reaction not in a model
    reaction.gene_reaction_rule = "(g1 or g2) and g3"
    assert reaction.gene_reaction_rule == "(g1 or g2) and g3"
    assert len(reaction.genes) == 3

    # Adding reaction with a GPR propagates to the model
    model.add_reactions([reaction])
    assert len(model.genes) == 3

    # Ensure the gene objects are the same in the model and reaction
    reaction_gene = list(reaction.genes)[0]
    model_gene = model.genes.get_by_id(reaction_gene.id)
    assert reaction_gene is model_gene

    # Test ability to handle uppercase AND/OR
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        reaction.gene_reaction_rule = "(b1 AND b2) OR (b3 and b4)"
        assert reaction.gene_reaction_rule == "(b1 and b2) or (b3 and b4)"
        assert len(reaction.genes) == 4

    # Ensure regular expressions correctly extract genes from malformed
    # GPR string
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        reaction.gene_reaction_rule = "(a1 or a2"
        assert len(reaction.genes) == 2
        reaction.gene_reaction_rule = "(forT or "
        assert len(reaction.genes) == 1


def test_gpr_modification(model):
    reaction = model.reactions.get_by_id("PGI")
    old_gene = list(reaction.genes)[0]
    new_gene = model.genes.get_by_id("s0001")

    # Add an existing 'gene' to the GPR
    reaction.gene_reaction_rule = 's0001'
    assert new_gene in reaction.genes
    assert reaction in new_gene.reactions

    # Remove old gene correctly
    assert old_gene not in reaction.genes
    assert reaction not in old_gene.reactions

    # Add a new 'gene' to the GPR
    reaction.gene_reaction_rule = 'fake_gene'
    assert model.genes.has_id("fake_gene")
    fake_gene = model.genes.get_by_id("fake_gene")
    assert fake_gene in reaction.genes
    assert reaction in fake_gene.reactions
    fake_gene.name = "foo_gene"
    assert reaction.gene_name_reaction_rule == fake_gene.name


def test_gene_knock_out(model):
    rxn = Reaction('rxn')
    rxn.add_metabolites({Metabolite('A'): -1, Metabolite('B'): 1})
    rxn.gene_reaction_rule = 'A2B1 or A2B2 and A2B3'
    assert hasattr(list(rxn.genes)[0], 'knock_out')
    model.add_reaction(rxn)
    with model:
        model.genes.A2B1.knock_out()
        assert not model.genes.A2B1.functional
        model.genes.A2B3.knock_out()
        assert not rxn.functional
    assert model.genes.A2B3.functional
    assert rxn.functional
    model.genes.A2B1.knock_out()
    assert not model.genes.A2B1.functional
    assert model.reactions.rxn.functional
    model.genes.A2B3.knock_out()
    assert not model.reactions.rxn.functional


def test_str():
    rxn = Reaction('rxn')
    rxn.add_metabolites({Metabolite('A'): -1, Metabolite('B'): 1})
    assert str(rxn) == 'rxn: A --> B'


def test_str_from_model(model):
    assert model.reactions[0].__str__().startswith('ACALD')


def test_add_metabolite_from_solved_model(solved_model):
    solution, model = solved_model
    pgi_reaction = model.reactions.PGI
    test_met = model.metabolites[0]
    pgi_reaction.add_metabolites({test_met: 42}, combine=False)
    assert pgi_reaction.metabolites[test_met] == 42.0
    assert model.constraints[
               test_met.id].expression.as_coefficients_dict()[
               pgi_reaction.forward_variable] == 42.0
    assert model.constraints[
               test_met.id].expression.as_coefficients_dict()[
               pgi_reaction.reverse_variable] == -42.0

    pgi_reaction.add_metabolites({test_met: -10}, combine=True)
    assert pgi_reaction.metabolites[test_met] == 32.0
    assert model.constraints[
               test_met.id].expression.as_coefficients_dict()[
               pgi_reaction.forward_variable] == 32.0
    assert model.constraints[
               test_met.id].expression.as_coefficients_dict()[
               pgi_reaction.reverse_variable] == -32.0

    pgi_reaction.add_metabolites({test_met: 0}, combine=False)
    with pytest.raises(KeyError):
        pgi_reaction.metabolites[test_met]
    assert model.constraints[
               test_met.id].expression.as_coefficients_dict()[
               pgi_reaction.forward_variable] == 0
    assert model.constraints[
               test_met.id].expression.as_coefficients_dict()[
               pgi_reaction.reverse_variable] == 0


@pytest.mark.parametrize("solver", stable_optlang)
def test_add_metabolite_benchmark(model, benchmark, solver):
    reaction = model.reactions.get_by_id("PGI")
    many_metabolites = dict((m, 1) for m in model.metabolites[0:50])

    def add_remove_metabolite():
        reaction.add_metabolites(many_metabolites)
        if not getattr(model, 'solver', None):
            solver_dict[solver].create_problem(model)
        for m, c in many_metabolites.items():
            try:
                reaction.subtract_metabolites(
                    {m: reaction.get_coefficient(m)})
            except KeyError:
                pass

    benchmark(add_remove_metabolite)


def test_add_metabolite(model):
    with pytest.raises(ValueError):
        model.add_metabolites(Metabolite())
    with model:
        with model:
            reaction = model.reactions.get_by_id("PGI")
            reaction.add_metabolites({model.metabolites[0]: 1})
            assert model.metabolites[0] in reaction._metabolites
            fake_metabolite = Metabolite("fake")
            reaction.add_metabolites({fake_metabolite: 1})
            assert fake_metabolite in reaction._metabolites
            assert model.metabolites.has_id("fake")
            assert model.metabolites.get_by_id("fake") is fake_metabolite
            assert len(model._contexts[0]._history) == 0

    assert fake_metabolite._model is None
    assert fake_metabolite not in reaction._metabolites
    assert "fake" not in model.metabolites

    # Test adding by string
    with model:
        reaction.add_metabolites({"g6p_c": -1})  # already in reaction
        assert reaction._metabolites[
            model.metabolites.get_by_id("g6p_c")] == -2
        reaction.add_metabolites({"h_c": 1})
        assert reaction._metabolites[
            model.metabolites.get_by_id("h_c")] == 1
        with pytest.raises(KeyError):
            reaction.add_metabolites({"missing": 1})

    assert reaction._metabolites[
        model.metabolites.get_by_id("g6p_c")] == -1
    assert model.metabolites.h_c not in reaction._metabolites

    # Test combine=False
    reaction = model.reactions.get_by_id("ATPM")
    old_stoich = reaction._metabolites[
        model.metabolites.get_by_id("h2o_c")]
    with model:
        reaction.add_metabolites({'h2o_c': 2.5}, combine=False)
        assert reaction._metabolites[
            model.metabolites.get_by_id("h2o_c")] == 2.5

    assert reaction._metabolites[
        model.metabolites.get_by_id("h2o_c")] == old_stoich

    # Test adding to a new Reaction
    reaction = Reaction("test")
    assert len(reaction._metabolites) == 0
    reaction.add_metabolites({Metabolite("test_met"): -1})
    assert len(reaction._metabolites) == 1


@pytest.mark.parametrize("solver", stable_optlang)
def test_subtract_metabolite_benchmark(model, benchmark, solver):
    benchmark(test_subtract_metabolite, model, solver)


@pytest.mark.parametrize("solver", stable_optlang)
def test_subtract_metabolite(model, solver):
    reaction = model.reactions.get_by_id("PGI")
    reaction.subtract_metabolites(reaction.metabolites)
    if not getattr(model, 'solver', None):
        solver_dict[solver].create_problem(model)
        assert len(reaction.metabolites) == 0


def test_mass_balance(model):
    reaction = model.reactions.get_by_id("PGI")
    # Should be balanced now
    assert len(reaction.check_mass_balance()) == 0
    # Should not be balanced after adding a hydrogen
    reaction.add_metabolites({model.metabolites.get_by_id("h_c"): 1})
    imbalance = reaction.check_mass_balance()
    assert "H" in imbalance
    assert imbalance["H"] == 1


def test_build_from_string(model):
    m = len(model.metabolites)
    pgi = model.reactions.get_by_id("PGI")
    old_bounds = pgi.bounds

    with model:
        pgi.reaction = "g6p_c --> f6p_c"
        assert pgi.lower_bound == 0

    assert pgi.bounds == old_bounds

    pgi.bounds = (0, 1000)
    assert pgi.bounds == (0, 1000)
    assert not pgi.reversibility
    pgi.reaction = "g6p_c <== f6p_c"
    assert pgi.upper_bound == 0
    assert pgi.reaction.strip() == "g6p_c <-- f6p_c"
    pgi.reaction = "g6p_c --> f6p_c + h2o_c"
    assert model.metabolites.h2o_c, pgi._metabolites

    with model:
        pgi.build_reaction_from_string("g6p_c --> f6p_c + foo",
                                       verbose=False)
        assert model.metabolites.h2o_c not in pgi._metabolites
        assert "foo" in model.metabolites
        assert model.metabolites.foo in pgi._metabolites
        assert len(model.metabolites) == m + 1

    assert model.metabolites.h2o_c in pgi._metabolites
    assert "foo" not in model.metabolites
    with pytest.raises(AttributeError):
        assert model.metabolites.foo
        assert len(model.metabolites) == m


def test_bounds_setter(model):
    rxn = model.reactions.get_by_id("PGI")
    with pytest.raises(ValueError):
        rxn.bounds = (1, 0)


def test_copy(model):
    PGI = model.reactions.PGI
    copied = PGI.copy()
    assert PGI is not copied
    assert PGI._model is model
    assert copied._model is not model
    # The copy should refer to different metabolites and genes
    for met in copied.metabolites:
        assert met is not model.metabolites.get_by_id(met.id)
        assert met.model is not model
    for gene in copied.genes:
        assert gene is not model.genes.get_by_id(gene.id)
        assert gene.model is not model


def test_iadd(model):
    PGI = model.reactions.PGI
    EX_h2o = model.reactions.EX_h2o_e
    original_PGI_gpr = PGI.gene_reaction_rule
    PGI += EX_h2o
    assert PGI.gene_reaction_rule == original_PGI_gpr
    assert PGI.metabolites[model.metabolites.h2o_e] == -1.0
    # Original should not change
    assert EX_h2o.gene_reaction_rule == ''
    assert EX_h2o.metabolites[model.metabolites.h2o_e] == -1.0
    # Add a reaction not in the model
    new_reaction = Reaction("test")
    new_reaction.add_metabolites({Metabolite("A"): -1, Metabolite("B"): 1})
    PGI += new_reaction
    assert PGI.gene_reaction_rule == original_PGI_gpr
    assert len(PGI.gene_reaction_rule) == 5
    # And vice versa
    new_reaction += PGI
    assert len(new_reaction.metabolites) == 5  # not
    assert len(new_reaction.genes) == 1
    assert new_reaction.gene_reaction_rule == original_PGI_gpr
    # Combine two GPRs
    model.reactions.ACKr += model.reactions.ACONTa
    expected_rule = '(b2296 or b3115 or b1849) and (b0118 or b1276)'
    assert model.reactions.ACKr.gene_reaction_rule == expected_rule
    assert len(model.reactions.ACKr.genes) == 5


def test_add(model):
    # Not in place addition should work on a copy
    new = model.reactions.PGI + model.reactions.EX_h2o_e
    assert new._model is not model
    assert len(new.metabolites) == 3
    # The copy should refer to different metabolites and genes
    # This currently fails because add_metabolites does not copy.
    # Should that be changed?
    # for met in new.metabolites:
    #    assert met is not model.metabolites.get_by_id(met.id)
    #    assert met.model is not model
    for gene in new.genes:
        assert gene is not model.genes.get_by_id(gene.id)
        assert gene.model is not model


def test_radd(model):
    new = sum([model.reactions.PGI, model.reactions.EX_h2o_e])
    assert new._model is not model
    assert len(new.metabolites) == 3


def test_mul(model):
    new = model.reactions.PGI * 2
    assert set(new.metabolites.values()) == {-2, 2}


def test_sub(model):
    new = model.reactions.PGI - model.reactions.EX_h2o_e
    assert new._model is not model
    assert len(new.metabolites) == 3


def test_removal_from_model_retains_bounds(model):
    model_cp = model.copy()
    reaction = model_cp.reactions.ACALD
    assert reaction.model == model_cp
    assert reaction.lower_bound == -1000.0
    assert reaction.upper_bound == 1000.0
    assert reaction._lower_bound == -1000.0
    assert reaction._upper_bound == 1000.0
    model_cp.remove_reactions([reaction])
    assert reaction.model is None
    assert reaction.lower_bound == -1000.0
    assert reaction.upper_bound == 1000.0
    assert reaction._lower_bound == -1000.0
    assert reaction._upper_bound == 1000.0


def test_set_bounds_scenario_1(model):
    acald_reaction = model.reactions.ACALD
    assert acald_reaction.lower_bound == -1000.
    assert acald_reaction.upper_bound == 1000.
    assert acald_reaction.forward_variable.lb == 0.
    assert acald_reaction.forward_variable.ub == 1000.
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 1000.
    acald_reaction.upper_bound = acald_reaction.lower_bound - 100
    assert acald_reaction.lower_bound == -1100.0
    assert acald_reaction.upper_bound == -1100.0
    assert acald_reaction.forward_variable.lb == 0
    assert acald_reaction.forward_variable.ub == 0
    assert acald_reaction.reverse_variable.lb == 1100.
    assert acald_reaction.reverse_variable.ub == 1100.
    acald_reaction.upper_bound = 100
    assert acald_reaction.lower_bound == -1100.0
    assert acald_reaction.upper_bound == 100
    assert acald_reaction.forward_variable.lb == 0
    assert acald_reaction.forward_variable.ub == 100
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 1100.0


def test_set_bounds_scenario_3(model):
    reac = model.reactions.ACALD
    reac.upper_bound = -10
    reac.lower_bound = -10
    assert reac.lower_bound == -10
    assert reac.upper_bound == -10
    reac.lower_bound = -9
    assert reac.lower_bound == -9
    assert reac.upper_bound == -9
    reac.lower_bound = 2
    assert reac.lower_bound == 2
    assert reac.upper_bound == 2
    reac.upper_bound = -10
    assert reac.lower_bound == -10
    assert reac.upper_bound == -10
    reac.upper_bound = -11
    assert reac.lower_bound == -11
    assert reac.upper_bound == -11
    reac.upper_bound = 2
    assert reac.lower_bound == -11
    assert reac.upper_bound == 2


def test_set_bounds_scenario_4(model):
    reac = model.reactions.ACALD
    reac.lower_bound = reac.upper_bound = 0
    reac.lower_bound = 2
    assert reac.lower_bound == 2
    assert reac.upper_bound == 2
    assert reac.forward_variable.lb == 2
    assert reac.forward_variable.ub == 2
    reac.knock_out()
    reac.upper_bound = -2
    assert reac.lower_bound == -2
    assert reac.upper_bound == -2
    assert reac.reverse_variable.lb == 2
    assert reac.reverse_variable.ub == 2


def test_set_upper_before_lower_bound_to_0(model):
    model.reactions.GAPD.upper_bound = 0
    model.reactions.GAPD.lower_bound = 0
    assert model.reactions.GAPD.lower_bound == 0
    assert model.reactions.GAPD.upper_bound == 0
    assert model.reactions.GAPD.forward_variable.lb == 0
    assert model.reactions.GAPD.forward_variable.ub == 0
    assert model.reactions.GAPD.reverse_variable.lb == 0
    assert model.reactions.GAPD.reverse_variable.ub == 0


def test_set_bounds_scenario_2(model):
    acald_reaction = model.reactions.ACALD
    assert acald_reaction.lower_bound == -1000.
    assert acald_reaction.upper_bound == 1000.
    assert acald_reaction.forward_variable.lb == 0.
    assert acald_reaction.forward_variable.ub == 1000.
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 1000.
    acald_reaction.lower_bound = acald_reaction.upper_bound + 100
    assert acald_reaction.lower_bound == 1100.0
    assert acald_reaction.upper_bound == 1100.0
    assert acald_reaction.forward_variable.lb == 1100.0
    assert acald_reaction.forward_variable.ub == 1100.0
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 0
    acald_reaction.lower_bound = -100
    assert acald_reaction.lower_bound == -100.
    assert acald_reaction.upper_bound == 1100.
    assert acald_reaction.forward_variable.lb == 0
    assert acald_reaction.forward_variable.ub == 1100.
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 100


def test_change_bounds(model):
    reac = model.reactions.ACALD
    reac.bounds = (2, 2)
    assert reac.lower_bound == 2
    assert reac.upper_bound == 2
    with model:
        reac.bounds = (5, 5)
        assert reac.lower_bound == 5
        assert reac.upper_bound == 5
    assert reac.lower_bound == 2
    assert reac.upper_bound == 2


def test_make_irreversible(model):
    acald_reaction = model.reactions.ACALD
    assert acald_reaction.lower_bound == -1000.
    assert acald_reaction.upper_bound == 1000.
    assert acald_reaction.forward_variable.lb == 0.
    assert acald_reaction.forward_variable.ub == 1000.
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 1000.
    acald_reaction.lower_bound = 0
    assert acald_reaction.lower_bound == 0
    assert acald_reaction.upper_bound == 1000.
    assert acald_reaction.forward_variable.lb == 0
    assert acald_reaction.forward_variable.ub == 1000.0
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 0
    acald_reaction.lower_bound = -100
    assert acald_reaction.lower_bound == -100.
    assert acald_reaction.upper_bound == 1000.
    assert acald_reaction.forward_variable.lb == 0
    assert acald_reaction.forward_variable.ub == 1000.
    assert acald_reaction.reverse_variable.lb == 0
    assert acald_reaction.reverse_variable.ub == 100


def test_make_reversible(model):
    pfk_reaction = model.reactions.PFK
    assert pfk_reaction.lower_bound == 0.
    assert pfk_reaction.upper_bound == 1000.
    assert pfk_reaction.forward_variable.lb == 0.
    assert pfk_reaction.forward_variable.ub == 1000.
    assert pfk_reaction.reverse_variable.lb == 0
    assert pfk_reaction.reverse_variable.ub == 0
    pfk_reaction.lower_bound = -100.
    assert pfk_reaction.lower_bound == -100.
    assert pfk_reaction.upper_bound == 1000.
    assert pfk_reaction.forward_variable.lb == 0
    assert pfk_reaction.forward_variable.ub == 1000.0
    assert pfk_reaction.reverse_variable.lb == 0
    assert pfk_reaction.reverse_variable.ub == 100.
    pfk_reaction.lower_bound = 0
    assert pfk_reaction.lower_bound == 0
    assert pfk_reaction.upper_bound == 1000.
    assert pfk_reaction.forward_variable.lb == 0
    assert pfk_reaction.forward_variable.ub == 1000.
    assert pfk_reaction.reverse_variable.lb == 0
    assert pfk_reaction.reverse_variable.ub == 0


def test_make_irreversible_irreversible_to_the_other_side(model):
    pfk_reaction = model.reactions.PFK
    assert pfk_reaction.lower_bound == 0.
    assert pfk_reaction.upper_bound == 1000.
    assert pfk_reaction.forward_variable.lb == 0.
    assert pfk_reaction.forward_variable.ub == 1000.
    assert pfk_reaction.reverse_variable.lb == 0
    assert pfk_reaction.reverse_variable.ub == 0
    pfk_reaction.upper_bound = -100.
    assert pfk_reaction.forward_variable.lb == 0
    assert pfk_reaction.forward_variable.ub == 0
    assert pfk_reaction.reverse_variable.lb == 100
    assert pfk_reaction.reverse_variable.ub == 100
    pfk_reaction.lower_bound = -1000.
    assert pfk_reaction.lower_bound == -1000.
    assert pfk_reaction.upper_bound == -100.
    assert pfk_reaction.forward_variable.lb == 0
    assert pfk_reaction.forward_variable.ub == 0
    assert pfk_reaction.reverse_variable.lb == 100
    assert pfk_reaction.reverse_variable.ub == 1000.


def test_make_lhs_irreversible_reversible(model):
    rxn = Reaction('test')
    rxn.add_metabolites(
        {model.metabolites[0]: -1., model.metabolites[1]: 1.})
    rxn.lower_bound = -1000.
    rxn.upper_bound = -100
    model.add_reaction(rxn)
    assert rxn.lower_bound == -1000.
    assert rxn.upper_bound == -100.
    assert rxn.forward_variable.lb == 0.
    assert rxn.forward_variable.ub == 0.
    assert rxn.reverse_variable.lb == 100.
    assert rxn.reverse_variable.ub == 1000.
    rxn.upper_bound = 666.
    assert rxn.lower_bound == -1000.
    assert rxn.upper_bound == 666.
    assert rxn.forward_variable.lb == 0.
    assert rxn.forward_variable.ub == 666
    assert rxn.reverse_variable.lb == 0.
    assert rxn.reverse_variable.ub == 1000.


def test_model_less_reaction(model):
    model.slim_optimize()
    for reaction in model.reactions:
        assert isinstance(reaction.flux, float)
        assert isinstance(reaction.reduced_cost, float)
    for reaction in model.reactions:
        model.remove_reactions([reaction])
        with pytest.raises(RuntimeError):
            reaction.flux
        with pytest.raises(RuntimeError):
            reaction.reduced_cost


def test_knockout(model):
    original_bounds = dict()
    for reaction in model.reactions:
        original_bounds[reaction.id] = (
            reaction.lower_bound, reaction.upper_bound)
        reaction.knock_out()
        assert reaction.lower_bound == 0
        assert reaction.upper_bound == 0
    for k, (lb, ub) in six.iteritems(original_bounds):
        model.reactions.get_by_id(k).lower_bound = lb
        model.reactions.get_by_id(k).upper_bound = ub
    for reaction in model.reactions:
        assert reaction.lower_bound == original_bounds[reaction.id][0]
        assert reaction.upper_bound == original_bounds[reaction.id][1]
    with model:
        for reaction in model.reactions:
            original_bounds[reaction.id] = (
                reaction.lower_bound, reaction.upper_bound)
            reaction.knock_out()
            assert reaction.lower_bound == 0
            assert reaction.upper_bound == 0
    for reaction in model.reactions:
        assert reaction.lower_bound == original_bounds[reaction.id][0]
        assert reaction.upper_bound == original_bounds[reaction.id][1]


def test_reaction_without_model():
    r = Reaction('blub')
    assert r.flux_expression is None
    assert r.forward_variable is None
    assert r.reverse_variable is None


def test_weird_left_to_right_reaction_issue(tiny_toy_model):
    d1 = tiny_toy_model.reactions.get_by_id('ex1')
    assert not d1.reversibility
    assert d1.lower_bound == -1000
    assert d1._lower_bound == -1000
    assert d1.upper_bound == 0
    assert d1._upper_bound == 0
    with tiny_toy_model:
        d1.knock_out()
        assert d1.lower_bound == 0
        assert d1._lower_bound == 0
        assert d1.upper_bound == 0
        assert d1._upper_bound == 0
    assert d1.lower_bound == -1000
    assert d1._lower_bound == -1000
    assert d1.upper_bound == 0
    assert d1._upper_bound == 0


def test_one_left_to_right_reaction_set_positive_ub(tiny_toy_model):
    d1 = tiny_toy_model.reactions.get_by_id('ex1')
    assert d1.reverse_variable.lb == 0
    assert d1.reverse_variable.ub == 1000
    assert d1._lower_bound == -1000
    assert d1.lower_bound == -1000
    assert d1._upper_bound == 0
    assert d1.upper_bound == 0
    assert d1.forward_variable.lb == 0
    assert d1.forward_variable.ub == 0
    d1.upper_bound = .1
    assert d1.forward_variable.lb == 0
    assert d1.forward_variable.ub == .1
    assert d1.reverse_variable.lb == 0
    assert d1.reverse_variable.ub == 1000
    assert d1._lower_bound == -1000
    assert d1.upper_bound == .1
    assert d1._lower_bound == -1000
    assert d1.upper_bound == .1


def test_irrev_reaction_set_negative_lb(model):
    assert not model.reactions.PFK.reversibility
    assert model.reactions.PFK.lower_bound == 0
    assert model.reactions.PFK.upper_bound == 1000.0
    assert model.reactions.PFK.forward_variable.lb == 0
    assert model.reactions.PFK.forward_variable.ub == 1000.0
    assert model.reactions.PFK.reverse_variable.lb == 0
    assert model.reactions.PFK.reverse_variable.ub == 0
    model.reactions.PFK.lower_bound = -1000
    assert model.reactions.PFK.lower_bound == -1000
    assert model.reactions.PFK.upper_bound == 1000.0
    assert model.reactions.PFK.forward_variable.lb == 0
    assert model.reactions.PFK.forward_variable.ub == 1000.0
    assert model.reactions.PFK.reverse_variable.lb == 0
    assert model.reactions.PFK.reverse_variable.ub == 1000


def test_twist_irrev_right_to_left_reaction_to_left_to_right(model):
    assert not model.reactions.PFK.reversibility
    assert model.reactions.PFK.lower_bound == 0
    assert model.reactions.PFK.upper_bound == 1000.0
    assert model.reactions.PFK.forward_variable.lb == 0
    assert model.reactions.PFK.forward_variable.ub == 1000.0
    assert model.reactions.PFK.reverse_variable.lb == 0
    assert model.reactions.PFK.reverse_variable.ub == 0
    model.reactions.PFK.lower_bound = -1000
    model.reactions.PFK.upper_bound = 0
    assert model.reactions.PFK.lower_bound == -1000
    assert model.reactions.PFK.upper_bound == 0
    assert model.reactions.PFK.forward_variable.lb == 0
    assert model.reactions.PFK.forward_variable.ub == 0
    assert model.reactions.PFK.reverse_variable.lb == 0
    assert model.reactions.PFK.reverse_variable.ub == 1000


def test_set_lb_higher_than_ub_sets_ub_to_new_lb(model):
    for reaction in model.reactions:
        assert reaction.lower_bound <= reaction.upper_bound
        reaction.lower_bound = reaction.upper_bound + 100
        assert reaction.lower_bound == reaction.upper_bound


def test_set_ub_lower_than_lb_sets_lb_to_new_ub(model):
    for reaction in model.reactions:
        assert reaction.lower_bound <= reaction.upper_bound
        reaction.upper_bound = reaction.lower_bound - 100
        assert reaction.lower_bound == reaction.upper_bound


def test_add_metabolites_combine_true(model):
    test_metabolite = Metabolite('test')
    for reaction in model.reactions:
        reaction.add_metabolites({test_metabolite: -66}, combine=True)
        assert reaction.metabolites[test_metabolite] == -66
        assert model.constraints['test'].get_linear_coefficients(
            [reaction.forward_variable])[reaction.forward_variable] == -66
        assert model.constraints['test'].get_linear_coefficients(
            [reaction.reverse_variable])[reaction.reverse_variable] == 66
        already_included_metabolite = \
            list(reaction.metabolites.keys())[0]
        previous_coefficient = reaction.get_coefficient(
            already_included_metabolite.id)
        reaction.add_metabolites({already_included_metabolite: 10},
                                 combine=True)
        new_coefficient = previous_coefficient + 10
        assert reaction.metabolites[
                   already_included_metabolite] == new_coefficient
        assert (model.constraints[
                already_included_metabolite.id].get_linear_coefficients(
                [reaction.forward_variable])[reaction.forward_variable] ==
                new_coefficient)
        assert (model.constraints[
                already_included_metabolite.id].get_linear_coefficients(
                [reaction.reverse_variable])[
                    reaction.reverse_variable] == -new_coefficient)


@pytest.mark.xfail(reason='non-deterministic test')
def test_add_metabolites_combine_false(model):
    test_metabolite = Metabolite('test')
    for reaction in model.reactions:
        reaction.add_metabolites({test_metabolite: -66}, combine=False)
        assert reaction.metabolites[test_metabolite] == -66
        assert model.constraints['test'].expression.has(
            -66. * reaction.forward_variable)
        assert model.constraints['test'].expression.has(
            66. * reaction.reverse_variable)
        already_included_metabolite = \
            list(reaction.metabolites.keys())[0]
        reaction.add_metabolites({already_included_metabolite: 10},
                                 combine=False)
        assert reaction.metabolites[already_included_metabolite] == 10
        assert model.constraints[
            already_included_metabolite.id].expression.has(
            10 * reaction.forward_variable)
        assert model.constraints[
            already_included_metabolite.id].expression.has(
            -10 * reaction.reverse_variable)


def test_reaction_imul(model):
    with model:
        model.reactions.EX_glc__D_e *= 100
        assert model.constraints.glc__D_e.expression.coeff(
            model.variables.EX_glc__D_e) == -100.0
        assert model.reactions.EX_glc__D_e.reaction == \
            '100.0 glc__D_e <=> '

    assert model.constraints.glc__D_e.expression.coeff(
        model.variables.EX_glc__D_e) == -1.0
    assert model.reactions.EX_glc__D_e.reaction == 'glc__D_e <=> '

    with model:
        model.reactions.EX_glc__D_e *= -2
        assert model.reactions.EX_glc__D_e.bounds == (-1000.0, 10.0)
        assert model.reactions.EX_glc__D_e.reaction == ' <=> 2.0 glc__D_e'

    assert model.reactions.EX_glc__D_e.bounds == (-10, 1000.0)
    assert model.reactions.EX_glc__D_e.reaction == 'glc__D_e <=> '


# def test_pop(model):
#     pgi = model.reactions.PGI
#     g6p = model.metabolites.get_by_id("g6p_c")
#     f6p = model.metabolites.get_by_id("f6p_c")
#     g6p_expr = model.solver.constraints["g6p_c"].expression
#     g6p_coef = pgi.pop("g6p_c")
#     assert g6p not in pgi.metabolites
#     actual = model.solver.constraints[
#         "g6p_c"].expression.as_coefficients_dict()
#     expected = (g6p_expr - g6p_coef * pgi.flux_expression
#                 ).as_coefficients_dict()
#     assert actual == expected
#     assert pgi.metabolites[f6p] == 1
#
#     f6p_expr = model.solver.constraints["f6p_c"].expression
#     f6p_coef = pgi.pop(f6p)
#     assert f6p not in pgi.metabolites
#     assert model.solver.constraints[
#                "f6p_c"].expression.as_coefficients_dict() == (
#                f6p_expr - f6p_coef * pgi.flux_expression
#            ).as_coefficients_dict()


def test_remove_from_model(model):
    pgi = model.reactions.PGI
    g6p = model.metabolites.g6p_c
    pgi_flux = model.optimize().fluxes['PGI']
    assert abs(pgi_flux) > 1E-6

    with model:
        pgi.remove_from_model()
        assert pgi.model is None
        assert "PGI" not in model.reactions
        assert pgi.id not in model.variables
        assert pgi.reverse_id not in model.variables
        assert pgi not in g6p.reactions
        model.optimize()

    assert "PGI" in model.reactions
    assert pgi.id in model.variables
    assert pgi.reverse_id in model.variables
    assert pgi.forward_variable.problem is model.solver
    assert pgi in g6p.reactions
    assert g6p in pgi.metabolites
    assert np.isclose(pgi_flux, model.optimize().fluxes['PGI'])


def test_change_id_is_reflected_in_solver(model):
    for i, reaction in enumerate(model.reactions):
        old_reaction_id = reaction.id
        assert model.variables[
                   old_reaction_id].name == old_reaction_id
        assert old_reaction_id in model.variables
        new_reaction_id = reaction.id + '_' + str(i)
        reaction.id = new_reaction_id
        assert reaction.id == new_reaction_id
        assert not (old_reaction_id in model.variables)
        assert reaction.id in model.variables
        assert reaction.reverse_id in model.variables
        name = model.variables[reaction.id].name
        assert name == reaction.id


def test_repr_html_(model):
    assert '<table>' in model.reactions[0]._repr_html_()



# -*- coding: utf-8 -*-

"""Define the global configuration."""

from __future__ import absolute_import

import logging
import types
from multiprocessing import cpu_count
from warnings import warn

from six import string_types, with_metaclass

from cobra.core.singleton import Singleton
from cobra.exceptions import SolverNotFound
from cobra.util.solver import interface_to_str, solvers


__all__ = ("Configuration",)


LOGGER = logging.getLogger(__name__)


class BaseConfiguration(object):
    """
    Define global configuration value that will be honored by cobra functions.

    This object sets default values for the modifiable attributes like
    default solver, reaction bounds etc.

    Attributes
    ----------
    solver : {"glpk", "cplex", "gurobi"}
        The default solver for new models. The solver choices are the ones
        provided by `optlang` and solvers installed in your environment.
    lower_bound : float
        The standard lower bound for reversible reactions (default -1000).
    upper_bound : float
        The standard upper bound for all reactions (default 1000).
    bounds : tuple of floats
        The default reaction bounds for newly created reactions. The bounds
        are in the form of lower_bound, upper_bound (default -1000.0, 1000.0).
    processes : int
        A default number of processes to use where multiprocessing is
        possible. The default number corresponds to the number of available
        cores (hyperthreads).

    """

    def __init__(self):
        self._solver = None
        self.lower_bound = None
        self.upper_bound = None
        # Set the default solver from a preferred order.
        for name in ["gurobi", "cplex", "glpk"]:
            try:
                self.solver = name
            except SolverNotFound:
                continue
            else:
                break
        self.bounds = -1000.0, 1000.0
        try:
            self.processes = cpu_count()
        except NotImplementedError:
            LOGGER.warning(
                "The number of cores could not be detected - assuming 1.")
            self.processes = 1

    @property
    def solver(self):
        return self._solver

    @solver.setter
    def solver(self, value):
        not_valid_interface = SolverNotFound(
            "'{}' is not a valid solver interface. Pick one from {}.".format(
                value, ", ".join(list(solvers))))
        if isinstance(value, string_types):
            if value not in solvers:
                raise not_valid_interface
            interface = solvers[value]
        elif isinstance(value, types.ModuleType) and hasattr(value, 'Model'):
            interface = value
        else:
            raise not_valid_interface
        self._solver = interface

    @property
    def bounds(self):
        return self.lower_bound, self.upper_bound

    @bounds.setter
    def bounds(self, bounds):
        # TODO: We should consider allowing `None` for free bounds.
        assert bounds[0] <= bounds[1]
        self.lower_bound = bounds[0]
        self.upper_bound = bounds[1]

    def __repr__(self):
        return "solver: {}\nlower_bound: {}\nupper_bound: {}\n" \
            "processes: {}".format(
                interface_to_str(self.solver), self.lower_bound,
                self.upper_bound, self.processes)


class Configuration(with_metaclass(Singleton, BaseConfiguration)):
    """Define the configuration to be singleton based."""

    pass
